"""Lower-level plotting tools.

Routines that may be of use to users wishing for more fine-grained control may
wish to use.

- :func:`anesthetic.plot.make_1d_axes`
- :func:`anesthetic.plot.make_2d_axes`

to create a set of axes and legend proxies.

"""
from packaging import version
import numpy as np
from pandas import Series, DataFrame
import matplotlib.pyplot as plt
from scipy.stats import gaussian_kde
from scipy.special import erf
from matplotlib.gridspec import GridSpec, GridSpecFromSubplotSpec
from matplotlib.axes import Axes
import matplotlib.cbook as cbook
import matplotlib.lines as mlines
from matplotlib.ticker import MaxNLocator, AutoMinorLocator, LogLocator
from matplotlib.colors import LinearSegmentedColormap
from matplotlib.transforms import Affine2D
from anesthetic.utils import nest_level
from anesthetic.utils import (sample_compression_1d, quantile,
                              triangular_sample_compression_2d,
                              iso_probability_contours,
                              match_contour_to_contourf, histogram_bin_edges)
from anesthetic.boundary import cut_and_normalise_gaussian


class AxesSeries(Series):
    """Anesthetic's axes version of :class:`pandas.Series`.

    Parameters
    ----------
    index : list(str)
        Parameters to be placed on the y-axes.
    fig : :class:`matplotlib.figure.Figure`
    ncol : int
        Number of axes columns. Decides after how many axes the AxesSeries is
        split to continue in a new row.
    labels : dict(str:str), optional
        Dictionary mapping params to plot labels.
        Default: params
    logx : list(str), optional
        List of parameters to be plotted on a log scale.
    gridspec_kw : dict, optional
        Dict with keywords passed to the :class:`matplotlib.gridspec.GridSpec`
        constructor used to create the grid the subplots are placed on.
    subplot_spec : :class:`matplotlib.gridspec.GridSpec`, default=None
        GridSpec instance to plot array as part of a subfigure.

    Methods
    -------
    set_xlabels:
        Set the labels for the x-axes.
    tick_params:
        Set tick parameters across all axes.

    """

    _metadata = ['_logx']
    _logx = []

    def __init__(self, data=None, index=None, fig=None, ncol=None, labels=None,
                 logx=None, gridspec_kw=None, subplot_spec=None,
                 *args, **kwargs):
        if data is None and index is not None:
            data = self.axes_series(index=index, fig=fig, ncol=ncol,
                                    gridspec_kw=gridspec_kw,
                                    subplot_spec=subplot_spec)
            self._set_xlabels(axes=data, labels=labels)
        super().__init__(data=data, index=index, *args, **kwargs)
        if logx is None:
            self._logx = []
        else:
            self._logx = logx
            self._set_xscale()

    @property
    def _constructor(self):
        return AxesSeries

    @property
    def _constructor_expanddim(self):
        return AxesDataFrame

    @staticmethod
    def axes_series(index, fig, ncol=None, gridspec_kw=None,
                    subplot_spec=None):
        """Set up subplots for :class:`AxesSeries`."""
        axes = Series(np.full(np.shape(index), None), index=index)
        if fig is None:
            fig = plt.figure()
        if ncol is None:
            ncol = int(np.ceil(np.sqrt(axes.index.size)))
        ncol = ncol
        nrow = int(np.ceil(axes.index.size / ncol))
        if gridspec_kw is None:
            gridspec_kw = {}
        wspace = gridspec_kw.pop('wspace', 0)
        if subplot_spec is None:
            gs = GridSpec(nrow, ncol, wspace=wspace, **gridspec_kw)
        else:
            gs = GridSpecFromSubplotSpec(nrow, ncol, wspace=wspace,
                                         subplot_spec=subplot_spec,
                                         **gridspec_kw)
        for p, g in zip(axes.index, gs):
            axes[p] = ax = fig.add_subplot(g)
            ax.set_yticks([])
        return axes

    def _set_xscale(self):
        for p, ax in self.items():
            if p in self._logx:
                ax.set_xscale('log')

    @staticmethod
    def _set_xlabels(axes, labels, **kwargs):
        if labels is None:
            labels = {}
        labels = {p: labels[p] if p in labels else p for p in axes.index}
        for p, ax in axes.items():
            ax.set_xlabel(labels[p], **kwargs)

    def set_xlabels(self, labels, **kwargs):
        """Set the labels for the x-axes.

        Parameters
        ----------
            labels : dict
                Dictionary of the axes labels.
            kwargs
                Any kwarg that can be passed to
                :meth:`matplotlib.axes.Axes.set_xlabel`.

        """
        self._set_xlabels(axes=self, labels=labels, **kwargs)

    def tick_params(self, *args, **kwargs):
        """Apply :meth:`matplotlib.axes.Axes.tick_params` across all axes."""
        for p, ax in self.items():
            ax.tick_params(*args, **kwargs)


class AxesDataFrame(DataFrame):
    """Anesthetic's axes version of :class:`pandas.DataFrame`.

    Parameters
    ----------
    index : list(str)
        Parameters to be placed on the y-axes.
    columns : list(str)
        Parameters to be placed on the x-axes.
    fig : :class:`matplotlib.figure.Figure`
    lower, diagonal, upper : bool, default=True
        Whether to create 2D marginalised plots above or below the
        diagonal, or to create a 1D marginalised plot on the diagonal.
    labels : dict(str:str), optional
        Dictionary mapping params to plot labels.
        Default: params
    ticks : str, default='inner'
        If 'outer', plot ticks only on the very left and very bottom.
        If 'inner', plot ticks also in inner subplots.
        If None, plot no ticks at all.
    logx, logy : list(str), optional
        Lists of parameters to be plotted on a log scale on the x-axis or
        y-axis, respectively.
    gridspec_kw : dict, optional
        Dict with keywords passed to the :class:`matplotlib.gridspec.GridSpec`
        constructor used to create the grid the subplots are placed on.
    subplot_spec : :class:`matplotlib.gridspec.GridSpec`, default=None
        GridSpec instance to plot array as part of a subfigure.

    Methods
    -------
    axlines:
        Add vertical and horizontal lines across all axes.
    axspans:
        Add vertical and horizontal spans across all axes.
    scatter:
        Add scatter points across all axes.
    set_labels:
        Set the labels for the axes.
    set_margins:
        Set margins across all axes.
    tick_params:
        Set tick parameters across all axes.

    """

    _metadata = ['_logx', '_logy']
    _logx = []
    _logy = []

    def __init__(self, data=None, index=None, columns=None, fig=None,
                 lower=True, diagonal=True, upper=True, labels=None,
                 ticks='inner', logx=None, logy=None,
                 gridspec_kw=None, subplot_spec=None, *args, **kwargs):
        if data is None and index is not None and columns is not None:
            position = self._position_frame(index=index,
                                            columns=columns,
                                            lower=lower,
                                            diagonal=diagonal,
                                            upper=upper)
            data = self._axes_frame(position=position,
                                    fig=fig,
                                    gridspec_kw=gridspec_kw,
                                    subplot_spec=subplot_spec)
            self._set_labels(axes=data, labels=labels)
            index = data.index
            columns = data.columns
            self._tick_params(axes=data, direction=ticks, which='both')
        super().__init__(data=data,
                         index=index,
                         columns=columns,
                         *args, **kwargs)
        if logx is None:
            self._logx = []
        else:
            self._logx = logx
        if logy is None:
            self._logy = []
        else:
            self._logy = logy
        if self._logx or self._logy:
            self._set_scale()
        self.tick_params(axis='both', which='both', labelrotation=45,
                         labelsize='small')

    @property
    def _constructor(self):
        return AxesDataFrame

    @property
    def _constructor_sliced(self):
        return AxesSeries

    @staticmethod
    def _position_frame(index, columns, lower, diagonal, upper):
        """Compute positions with lower=-1, diagonal=0, upper=+1."""
        data = np.full((np.size(index), np.size(columns)), None)
        position = DataFrame(data=data, index=index, columns=columns)
        all_params = list(columns) + list(index)
        for j, y in enumerate(index):
            for i, x in enumerate(columns):
                if all_params.index(x) < all_params.index(y):
                    if lower:
                        position.loc[y, x] = -1
                elif all_params.index(x) > all_params.index(y):
                    if upper:
                        position.loc[y, x] = +1
                elif diagonal:
                    position.loc[y, x] = 0
        return position

    @classmethod
    def _axes_frame(cls, position, fig, gridspec_kw=None, subplot_spec=None):
        """Set up subplots for :class:`AxesDataFrame`."""
        axes = position.copy()
        axes.dropna(axis=0, how='all', inplace=True)
        axes.dropna(axis=1, how='all', inplace=True)
        if axes.size == 0:
            return axes
        if fig is None:
            fig = plt.figure()
        if gridspec_kw is None:
            gridspec_kw = {}
        hspace = gridspec_kw.pop('hspace', 0)
        wspace = gridspec_kw.pop('wspace', 0)
        if subplot_spec is None:
            gs = GridSpec(*axes.shape, hspace=hspace, wspace=wspace,
                          **gridspec_kw)
        else:
            gs = GridSpecFromSubplotSpec(*axes.shape,
                                         hspace=hspace, wspace=wspace,
                                         subplot_spec=subplot_spec,
                                         **gridspec_kw)
        axes.loc[:, :] = None
        for j, y in enumerate(axes.index[::-1]):
            for i, x in enumerate(axes.columns):
                if position[x][y] is not None:
                    sx = list(axes[x].dropna())
                    sx = sx[0] if sx else None
                    sy = list(axes.T[y].dropna())
                    sy = sy[0] if sy else None
                    axes.loc[y, x] = fig.add_subplot(
                        gs[axes.index.size - 1 - j, i], sharex=sx, sharey=sy
                    )
                    if position[x][y] == 0:
                        axes[x][y].twin = axes[x][y].twinx()
                        axes[x][y].twin.set_yticks([])
                        cls._make_diagonal(axes[x][y])
                        axes[x][y].position = 'diagonal'
                        axes[x][y].twin.xaxis.set_major_locator(
                            MaxNLocator(3, prune='both'))
                        axes[x][y].twin.xaxis.set_minor_locator(
                            AutoMinorLocator(1))
                        axes[x][y].yaxis.set_major_locator(
                            MaxNLocator(3, prune='both'))
                        axes[x][y].yaxis.set_minor_locator(AutoMinorLocator(1))
                    else:
                        if position[x][y] == 1:
                            axes[x][y].position = 'upper'
                            cls._make_offdiagonal(axes[x][y])
                        elif position[x][y] == -1:
                            axes[x][y].position = 'lower'
                            cls._make_offdiagonal(axes[x][y])
                        axes[x][y].yaxis.set_major_locator(
                            MaxNLocator(3, prune='both'))
                        axes[x][y].yaxis.set_minor_locator(AutoMinorLocator(1))
                    axes[x][y].xaxis.set_major_locator(
                        MaxNLocator(3, prune='both'))
                    axes[x][y].xaxis.set_minor_locator(AutoMinorLocator(1))
        return axes

    @staticmethod
    def _make_diagonal(ax):
        """Link x and y axes limits."""

        class DiagonalAxes(type(ax)):
            def set_xlim(self, left=None, right=None, emit=True, auto=False,
                         xmin=None, xmax=None):
                if (self.get_xaxis().get_scale() ==
                        self.get_yaxis().get_scale()):
                    super().set_ylim(bottom=left, top=right, emit=True,
                                     auto=auto, ymin=xmin, ymax=xmax)
                return super().set_xlim(left=left, right=right, emit=emit,
                                        auto=auto, xmin=xmin, xmax=xmax)

            def set_ylim(self, bottom=None, top=None, emit=True, auto=False,
                         ymin=None, ymax=None):
                if (self.get_xaxis().get_scale() ==
                        self.get_yaxis().get_scale()):
                    super().set_xlim(left=bottom, right=top, emit=True,
                                     auto=auto, xmin=ymin, xmax=ymax)
                return super().set_ylim(bottom=bottom, top=top, emit=emit,
                                        auto=auto, ymin=ymin, ymax=ymax)

            def get_legend_handles_labels(self, *args, **kwargs):
                return self.twin.get_legend_handles_labels(*args, **kwargs)

            def legend(self, *args, **kwargs):
                return self.twin.legend(*args, **kwargs)

        ax.__class__ = DiagonalAxes

    @staticmethod
    def _make_offdiagonal(ax):
        """Linking x to y axes limits in triangle plots."""

        class OffDiagonalAxes(type(ax)):
            def set_xlim(self, left=None, right=None, emit=True, auto=False,
                         xmin=None, xmax=None):
                left, right = super().set_xlim(left=left, right=right,
                                               emit=emit,
                                               auto=auto, xmin=xmin, xmax=xmax)
                if emit:
                    self.callbacks.process('xlim_changed', self)
                    # Call all other x-axes that are shared with this one
                    for other in self._shared_axes['x'].get_siblings(self):
                        if other is not self:
                            other.set_xlim(left, right, emit=False, auto=auto)
                return left, right

            def set_ylim(self, bottom=None, top=None, emit=True, auto=False,
                         ymin=None, ymax=None):
                bottom, top = super().set_ylim(bottom=bottom, top=top,
                                               emit=emit,
                                               auto=auto, ymin=ymin, ymax=ymax)
                if emit:
                    self.callbacks.process('ylim_changed', self)
                    # Call all other y-axes that are shared with this one
                    for other in self._shared_axes['y'].get_siblings(self):
                        if other is not self:
                            other.set_ylim(bottom, top, emit=False, auto=auto)
                return bottom, top

        ax.__class__ = OffDiagonalAxes

    def _set_scale(self):
        for y, rows in self.iterrows():
            for x, ax in rows.items():
                if ax is not None:
                    if x in self._logx:
                        ax.set_xscale('log')
                    if y in self._logy:
                        ax.set_yscale('log')

    def _set_logticks(self):
        for y, rows in self.iterrows():
            for x, ax in rows.items():
                if ax is not None:
                    if x in self._logx:
                        ax.xaxis.set_major_locator(LogLocator(numticks=3))
                        if x != y:
                            ax.set_xlim(ax.dataLim.intervalx)
                    if y in self._logy:
                        ax.yaxis.set_major_locator(LogLocator(numticks=3))
                        if y != x:
                            ax.set_ylim(ax.dataLim.intervaly)

    @staticmethod
    def _set_labels(axes, labels, **kwargs):
        all_params = list(axes.columns) + list(axes.index)
        if labels is None:
            labels = {}
        labels = {p: labels[p] if p in labels else p for p in all_params}

        for y, axes_row in axes.iterrows():
            if axes_row.size:
                axes_row.dropna(inplace=True)
                axes_row.iloc[0].set_ylabel(labels[y], **kwargs)

        for x, axes_col in axes.items():
            if axes_col.size:
                axes_col.dropna(inplace=True)
                axes_col.iloc[-1].set_xlabel(labels[x], **kwargs)

    def set_labels(self, labels, **kwargs):
        """Set the labels for the axes.

        Parameters
        ----------
        labels : dict
            Dictionary of the axes labels.
        kwargs
            Any kwarg that can be passed to
            :meth:`matplotlib.axes.Axes.set_xlabel` or
            :meth:`matplotlib.axes.Axes.set_ylabel`.

        """
        self._set_labels(axes=self, labels=labels, **kwargs)

    @staticmethod
    def _tick_params(axes, direction='inner', **kwargs):
        if direction not in ['inner', 'outer', None]:
            raise ValueError("tick direction=%s was requested, but tick "
                             "direction can only be one of "
                             "['outer', 'inner', None]." % direction)

        # left and right ticks and labels
        for y, ax in axes.iterrows():
            ax_ = ax.dropna()
            if len(ax_) and direction == 'inner':
                for i, a in enumerate(ax_):
                    if i == 0:  # first column
                        tl = a.yaxis.majorTicks[0].tick1line.get_markersize()
                        if a.position == 'diagonal' and len(ax_) == 1:
                            a.tick_params('y', left=False, labelleft=False,
                                          **kwargs)
                        else:
                            a.tick_params('y', left=True, labelleft=True,
                                          **kwargs)
                    elif a.position == 'diagonal':  # not first column
                        a.tick_params('y', direction='out', length=tl/2,
                                      left=True, labelleft=False, **kwargs)
                    elif ax_.iloc[i-1].position == 'diagonal':  # next to diag
                        a.tick_params('y', direction='in', length=tl/2,
                                      left=True, labelleft=False, **kwargs)
                    else:  # not diagonal and not first column
                        a.tick_params('y', direction='inout',
                                      left=True, labelleft=False, **kwargs)
            elif len(ax_) and direction == 'outer':  # no inner ticks
                for a in ax_.iloc[1:]:
                    a.tick_params('y', left=False, labelleft=False, **kwargs)
            elif len(ax_) and direction is None:  # no ticks at all
                for a in ax_:
                    a.tick_params('y', left=False, right=False,
                                  labelleft=False, labelright=False, **kwargs)

        # bottom and top ticks and labels
        for x, ax in axes.items():
            ax_ = ax.dropna()
            if len(ax_):
                if direction == 'inner':
                    for i, a in enumerate(ax_):
                        if i == len(ax_) - 1:  # bottom row
                            a.tick_params('x', bottom=True, labelbottom=True,
                                          **kwargs)
                        else:  # not bottom row
                            a.tick_params('x', direction='inout',
                                          bottom=True, labelbottom=False,
                                          **kwargs)
                            if a.position == 'diagonal':
                                a.twin.tick_params('x', direction='inout',
                                                   bottom=True,
                                                   labelbottom=False, **kwargs)
                elif direction == 'outer':  # no inner ticks
                    for a in ax_.iloc[:-1]:
                        a.tick_params('x', bottom=False, labelbottom=False,
                                      **kwargs)
                elif direction is None:  # no ticks at all
                    for a in ax_:
                        a.tick_params('x', bottom=False, top=False,
                                      labelbottom=False, labeltop=False,
                                      **kwargs)

    def tick_params(self, *args, **kwargs):
        """Apply :meth:`matplotlib.axes.Axes.tick_params` across all axes."""
        for y, rows in self.iterrows():
            for x, ax in rows.items():
                if isinstance(ax, Axes):
                    ax.tick_params(*args, **kwargs)

    def set_margins(self, m):
        """Apply :meth:`matplotlib.axes.Axes.set_xmargin` across all axes."""
        unique_params = list(np.unique(list(self.index) + list(self.columns)))
        for y, rows in self.iterrows():
            for x, ax in rows.items():
                if isinstance(ax, Axes):
                    if x in unique_params:
                        xmin, xmax = ax.get_xlim()
                        xdelta = xmax - xmin
                        ax.set_xlim(xmin - m * xdelta, xmax + m * xdelta)
                        unique_params.remove(x)
                    if y in unique_params:
                        ymin, ymax = ax.get_ylim()
                        ydelta = ymax - ymin
                        ax.set_ylim(ymin - m * ydelta, ymax + m * ydelta)
                        unique_params.remove(y)

    def axlines(self, params, lower=True, diagonal=True, upper=True, **kwargs):
        """Add vertical and horizontal lines across all axes.

        Parameters
        ----------
        params : dict(array_like)
            Dictionary of parameter labels and desired values.
            Can provide more than one value per label.
        lower, diagonal, upper : bool, default=True
            Whether to plot the lines on the lower, diagonal,
            and/or upper triangle plots.
        kwargs
            Any kwarg that can be passed to
            :meth:`matplotlib.axes.Axes.axvline` or
            :meth:`matplotlib.axes.Axes.axhline`.

        """
        positions = ['lower' if lower else None,
                     'diagonal' if diagonal else None,
                     'upper' if upper else None]
        for y, rows in self.iterrows():
            for x, ax in rows.items():
                if ax is not None and ax.position in positions:
                    if x in params:
                        for v in np.atleast_1d(params[x]):
                            if ax.position == 'diagonal':
                                ax.twin.axvline(v, **kwargs)
                            else:
                                ax.axvline(v, **kwargs)
                    if y in params and ax.position != 'diagonal':
                        for v in np.atleast_1d(params[y]):
                            ax.axhline(v, **kwargs)

    def axspans(self, params, lower=True, diagonal=True, upper=True, **kwargs):
        """Add vertical and horizontal spans across all axes.

        Parameters
        ----------
        params : dict(array_like(2-tuple))
            Dictionary of parameter labels and desired value tuples.
            Can provide more than one value tuple per label.
            Each value tuple provides the min and max value for an axis span.
        lower, diagonal, upper : bool, default=True
            Whether to plot the spans on the lower, diagonal,
            and/or upper triangle plots.
        kwargs
            Any kwarg that can be passed to
            :meth:`matplotlib.axes.Axes.axvspan` or
            :meth:`matplotlib.axes.Axes.axhspan`.

        """
        kwargs = normalize_kwargs(kwargs)
        positions = ['lower' if lower else None,
                     'diagonal' if diagonal else None,
                     'upper' if upper else None]
        for y, rows in self.iterrows():
            for x, ax in rows.items():
                if ax is not None and ax.position in positions:
                    if x in params:
                        for vmin, vmax in np.atleast_2d(params[x]):
                            if ax.position == 'diagonal':
                                ax.twin.axvspan(vmin, vmax, **kwargs)
                            else:
                                ax.axvspan(vmin, vmax, **kwargs)
                    if y in params and ax.position != 'diagonal':
                        for vmin, vmax in np.atleast_2d(params[y]):
                            ax.axhspan(vmin, vmax, **kwargs)

    def scatter(self, params, lower=True, upper=True, **kwargs):
        """Add scatter points across all axes.

        Parameters
        ----------
        params : dict(array_like)
            Dictionary of parameter labels and desired values.
            Can provide more than one value per label, but length has to
            match for all parameter labels.
        lower, upper : bool, default=True
            Whether to plot the spans on the lower and/or upper triangle plots.
        kwargs
            Any kwarg that can be passed to
            :meth:`matplotlib.axes.Axes.scatter`.

        """
        positions = ['lower' if lower else None,
                     'upper' if upper else None]
        zorder = kwargs.pop('zorder', None)
        for y, rows in self.iterrows():
            for x, ax in rows.items():
                if ax is not None and ax.position in positions:
                    if x in params and y in params:
                        z = max([z.get_zorder() for z in ax.artists +
                                 ax.collections + ax.lines + ax.patches] + [0])
                        z = z+1 if zorder is None else zorder
                        ax.scatter(params[x], params[y], zorder=z, **kwargs)


def make_1d_axes(params, ncol=None, labels=None, logx=None,
                 gridspec_kw=None, subplot_spec=None, **fig_kw):
    """Create a set of axes for plotting 1D marginalised posteriors.

    Parameters
    ----------
    params : list(str)
        names of parameters.

    ncol : int
        Number of columns of the subplot grid.
        Default: ceil(sqrt(num_params))

    labels : dict(str:str), optional
        Dictionary mapping params to plot labels.
        Default: params

    logx : list(str), optional
        List of parameters to be plotted on a log scale.

    gridspec_kw : dict, optional
        Dict with keywords passed to the :class:`matplotlib.gridspec.GridSpec`
        constructor used to create the grid the subplots are placed on.

    subplot_spec : :class:`matplotlib.gridspec.GridSpec`, default=None
        GridSpec instance to plot array as part of a subfigure.

    **fig_kw
        All additional keyword arguments are passed to the
        :func:`matplotlib.pyplot.figure` call.
        Or directly pass the figure to plot on via the keyword 'fig'.

    Returns
    -------
    fig : :class:`matplotlib.figure.Figure`
        New or original (if supplied) figure object.

    axes : :class:`anesthetic.plot.AxesSeries`
        Pandas array of axes objects.

    """
    # TODO: remove this in version >= 2.1
    if 'tex' in fig_kw:
        raise NotImplementedError(
            "This is anesthetic 1.0 syntax. You need to update, e.g.\n"
            "make_1d_axes(..., tex=tex)     # anesthetic 1.0\n"
            "make_1d_axes(..., labels=tex)  # anesthetic 2.0"
            )
    fig = fig_kw.pop('fig') if 'fig' in fig_kw else plt.figure(**fig_kw)
    if np.array(params).ndim == 0:
        params = [params]
    axes = AxesSeries(index=params,
                      fig=fig,
                      ncol=ncol,
                      labels=labels,
                      logx=logx,
                      gridspec_kw=gridspec_kw,
                      subplot_spec=subplot_spec)
    if gridspec_kw is None:
        fig.tight_layout()
    return fig, axes


def make_2d_axes(params, labels=None, lower=True, diagonal=True, upper=True,
                 ticks='inner', logx=None, logy=None,
                 gridspec_kw=None, subplot_spec=None, **fig_kw):
    """Create a set of axes for plotting 2D marginalised posteriors.

    Parameters
    ----------
    params : lists of parameters
        Can be either:

        * ``list(str)`` if the x and y axes are the same
        * ``[list(str), list(str)]`` if the x and y axes are different

        Strings indicate the names of the parameters.

    labels : dict(str:str), optional
        Dictionary mapping params to plot labels.
        Default: params

    lower, diagonal, upper : logical, default=True
        Whether to create 2D marginalised plots above or below the
        diagonal, or to create a 1D marginalised plot on the diagonal.

    ticks : str, default='inner'
        Can be one of 'outer', 'inner', or None.

        * ``'outer'``: plot ticks only on the very left and very bottom.
        * ``'inner'``: plot ticks also in inner subplots.
        * ``None``: plot no ticks at all.

    logx, logy : list(str), optional
        Lists of parameters to be plotted on a log scale on the x-axis or
        y-axis, respectively.

    gridspec_kw : dict, optional
        Dict with keywords passed to the :class:`matplotlib.gridspec.GridSpec`
        constructor used to create the grid the subplots are placed on.

    subplot_spec : :class:`matplotlib.gridspec.GridSpec`, default=None
        GridSpec instance to plot array as part of a subfigure.

    **fig_kw
        All additional keyword arguments are passed to the
        :func:`matplotlib.pyplot.figure` call.
        Or directly pass the figure to plot on via the keyword 'fig'.

    Returns
    -------
    fig : :class:`matplotlib.figure.Figure`
        New or original (if supplied) figure object.

    axes : :class:`anesthetic.plot.AxesDataFrame`
        Pandas array of axes objects.

    """
    # TODO: remove this in version >= 2.1
    if 'tex' in fig_kw:
        raise NotImplementedError(
            "This is anesthetic 1.0 syntax. You need to update, e.g.\n"
            "make_2d_axes(..., tex=tex)     # anesthetic 1.0\n"
            "make_2d_axes(..., labels=tex)  # anesthetic 2.0"
            )
    fig = fig_kw.pop('fig') if 'fig' in fig_kw else plt.figure(**fig_kw)

    if nest_level(params) == 2:
        xparams, yparams = params
    else:
        xparams = yparams = params
    axes = AxesDataFrame(index=yparams,
                         columns=xparams,
                         fig=fig,
                         lower=lower,
                         diagonal=diagonal,
                         upper=upper,
                         labels=labels,
                         ticks=ticks,
                         logx=logx,
                         logy=logy,
                         gridspec_kw=gridspec_kw,
                         subplot_spec=subplot_spec)
    fig.align_labels([a for a in axes.values.flatten() if isinstance(a, Axes)])
    return fig, axes


def fastkde_plot_1d(ax, data, *args, **kwargs):
    """Plot a 1d marginalised distribution.

    This functions as a wrapper around :meth:`matplotlib.axes.Axes.plot`, with
    a kernel density estimation (KDE) computation provided by the package
    fastkde in-between. All remaining keyword arguments are passed onwards.

    Parameters
    ----------
    ax : :class:`matplotlib.axes.Axes`
        Axis object to plot on.

    data : np.array
        Uniformly weighted samples to generate kernel density estimator.

    xmin, xmax : float, default=None
        lower/upper prior bound

    levels : list
        Values at which to draw iso-probability lines.
        Optional,
        Default: [0.95, 0.68]

    q : int or float or tuple, default=5
        Quantile to determine the data range to be plotted.

        * ``0``: full data range, i.e. ``q=0`` --> quantile range (0, 1)
        * ``int``: q-sigma range, e.g. ``q=1`` --> quantile range (0.16, 0.84)
        * ``float``: percentile, e.g. ``q=0.8`` --> quantile range (0.1, 0.9)
        * ``tuple``: quantile range, e.g. (0.16, 0.84)

    facecolor : bool or string, default=False
        If set to True then the 1d plot will be shaded with the value of the
        ``color`` kwarg. Set to a string such as 'blue', 'k', 'r', 'C1' ect.
        to define the color of the shading directly.

    Returns
    -------
    lines : :class:`matplotlib.lines.Line2D`
        A list of line objects representing the plotted data (same as
        :meth:`matplotlib.axes.Axes.plot` command).

    """
    kwargs = normalize_kwargs(kwargs)
    if ax.get_xaxis().get_scale() == 'log':
        data = np.log10(data)
    xmin = kwargs.pop('xmin', None)
    xmax = kwargs.pop('xmax', None)
    levels = kwargs.pop('levels', [0.95, 0.68])
    density = kwargs.pop('density', False)

    cmap = kwargs.pop('cmap', None)
    color = kwargs.pop('color', (ax._get_lines.get_next_color()
                                 if cmap is None
                                 else plt.get_cmap(cmap)(0.68)))
    facecolor = kwargs.pop('facecolor', False)
    if 'edgecolor' in kwargs:
        edgecolor = kwargs.pop('edgecolor')
        if edgecolor:
            color = edgecolor
    else:
        edgecolor = color

    q = kwargs.pop('q', 5)
    q = quantile_plot_interval(q=q)

    try:
        from anesthetic.kde import fastkde_1d
        x, p, xmin, xmax = fastkde_1d(data, xmin, xmax)
    except ImportError:
        raise ImportError("You need to install fastkde to use fastkde")
    p /= p.max()
    i = ((x > quantile(x, q[0], p)) & (x < quantile(x, q[-1], p)))

    if version.parse(np.__version__) >= version.parse("2.0.0"):
        trapezoid = np.trapezoid
    else:
        trapezoid = np.trapz
    area = trapezoid(x=x[i], y=p[i]) if density else 1
    if ax.get_xaxis().get_scale() == 'log':
        x = 10**x
    ans = ax.plot(x[i], p[i]/area, color=color, *args, **kwargs)

    if facecolor and facecolor not in [None, 'None', 'none']:
        if facecolor is True:
            facecolor = color
        c = iso_probability_contours(p[i], contours=levels)
        cmap = basic_cmap(facecolor)
        fill = []
        for j in range(len(c)-1):
            fill.append(ax.fill_between(x[i], p[i], where=p[i] >= c[j],
                        color=cmap(c[j]), edgecolor=edgecolor))

        return ans, fill

    return ans


def kde_plot_1d(ax, data, *args, **kwargs):
    """Plot a 1d marginalised distribution.

    This functions as a wrapper around :meth:`matplotlib.axes.Axes.plot`, with
    a kernel density estimation computation provided by
    :class:`scipy.stats.gaussian_kde` in-between. All remaining keyword
    arguments are passed onwards.

    Parameters
    ----------
    ax : :class:`matplotlib.axes.Axes`
        Axis object to plot on.

    data : np.array
        Samples to generate kernel density estimator.

    weights : np.array, optional
        Sample weights.

    ncompress : int, str, default=False
        Degree of compression.

        * If ``False``: no compression.
        * If ``True``: compresses to the channel capacity, equivalent to
          ``ncompress='entropy'``.
        * If ``int``: desired number of samples after compression.
        * If ``str``: determine number from the Huggins-Roy family of
          effective samples in :func:`anesthetic.utils.neff`
          with ``beta=ncompress``.

    nplot_1d : int, default=100
        Number of plotting points to use.

    levels : list
        Values at which to draw iso-probability lines.
        Default: [0.95, 0.68]

    q : int or float or tuple, default=5
        Quantile to determine the data range to be plotted.

        * ``0``: full data range, i.e. ``q=0`` --> quantile range (0, 1)
        * ``int``: q-sigma range, e.g. ``q=1`` --> quantile range (0.16, 0.84)
        * ``float``: percentile, e.g. ``q=0.8`` --> quantile range (0.1, 0.9)
        * ``tuple``: quantile range, e.g. (0.16, 0.84)

    facecolor : bool or string, default=False
        If set to True then the 1d plot will be shaded with the value of the
        ``color`` kwarg. Set to a string such as 'blue', 'k', 'r', 'C1' ect.
        to define the color of the shading directly.

    bw_method : str, scalar or callable, optional
        Forwarded to :class:`scipy.stats.gaussian_kde`.

    beta : int, float, default = 1
        The value of beta used to calculate the number of effective samples

    Returns
    -------
    lines : :class:`matplotlib.lines.Line2D`
        A list of line objects representing the plotted data (same as
        :meth:`matplotlib.axes.Axes.plot` command).

    """
    kwargs = normalize_kwargs(kwargs)
    weights = kwargs.pop('weights', None)
    if weights is not None:
        data = data[weights != 0]
        weights = weights[weights != 0]
    if ax.get_xaxis().get_scale() == 'log':
        data = np.log10(data)

    ncompress = kwargs.pop('ncompress', False)
    nplot = kwargs.pop('nplot_1d', 100)
    bw_method = kwargs.pop('bw_method', None)
    levels = kwargs.pop('levels', [0.95, 0.68])
    density = kwargs.pop('density', False)

    cmap = kwargs.pop('cmap', None)
    color = kwargs.pop('color', (ax._get_lines.get_next_color()
                                 if cmap is None
                                 else plt.get_cmap(cmap)(0.68)))
    facecolor = kwargs.pop('facecolor', False)
    if 'edgecolor' in kwargs:
        edgecolor = kwargs.pop('edgecolor')
        if edgecolor:
            color = edgecolor
    else:
        edgecolor = color

    q = kwargs.pop('q', 5)
    q = quantile_plot_interval(q=q)
    xmin = quantile(data, q[0], weights)
    xmax = quantile(data, q[-1], weights)
    x = np.linspace(xmin, xmax, nplot)

    data_compressed, w = sample_compression_1d(data, weights, ncompress)
    kde = gaussian_kde(data_compressed, weights=w, bw_method=bw_method)

    p = kde(x)
    p /= p.max()
    bw = np.sqrt(kde.covariance[0, 0])
    pp = cut_and_normalise_gaussian(x, p, bw, xmin=data.min(), xmax=data.max())
    pp /= pp.max()
    if version.parse(np.__version__) >= version.parse("2.0.0"):
        trapezoid = np.trapezoid
    else:
        trapezoid = np.trapz
    area = trapezoid(x=x, y=pp) if density else 1
    if ax.get_xaxis().get_scale() == 'log':
        x = 10**x
    ans = ax.plot(x, pp/area, color=color, *args, **kwargs)

    if facecolor and facecolor not in [None, 'None', 'none']:
        if facecolor is True:
            facecolor = color
        c = iso_probability_contours(pp, contours=levels)
        cmap = basic_cmap(facecolor)
        fill = []
        for j in range(len(c)-1):
            fill.append(ax.fill_between(x, pp, where=pp >= c[j],
                        color=cmap(c[j]), edgecolor=edgecolor))

        ans = ans, fill

    if density:
        ax.set_ylim(bottom=0)
    else:
        ax.set_ylim(0, 1.1)

    return ans


def hist_plot_1d(ax, data, *args, **kwargs):
    """Plot a 1d histogram.

    This functions is a wrapper around :meth:`matplotlib.axes.Axes.hist`. All
    remaining keyword arguments are passed onwards.

    Parameters
    ----------
    ax : :class:`matplotlib.axes.Axes`
        Axis object to plot on.

    data : np.array
        Samples to generate histogram from

    weights : np.array, optional
        Sample weights.

    q : int or float or tuple, default=5
        Quantile to determine the data range to be plotted.

        * ``0``: full data range, i.e. ``q=0`` --> quantile range (0, 1)
        * ``int``: q-sigma range, e.g. ``q=1`` --> quantile range (0.16, 0.84)
        * ``float``: percentile, e.g. ``q=0.8`` --> quantile range (0.1, 0.9)
        * ``tuple``: quantile range, e.g. (0.16, 0.84)

    Returns
    -------
    patches : list or list of lists
        Silent list of individual patches used to create the histogram
        or list of such list if multiple input datasets.

    Other Parameters
    ----------------
    **kwargs : :meth:`matplotlib.axes.Axes.hist` properties

    """
    kwargs = normalize_kwargs(kwargs)
    weights = kwargs.pop('weights', None)
    bins = kwargs.pop('bins', 'fd')
    histtype = kwargs.pop('histtype', 'bar')
    density = kwargs.get('density', False)

    cmap = kwargs.pop('cmap', None)
    color = kwargs.pop('color', (ax._get_lines.get_next_color()
                                 if cmap is None
                                 else plt.get_cmap(cmap)(0.68)))

    q = kwargs.pop('q', 5)
    q = quantile_plot_interval(q=q)
    if ax.get_xaxis().get_scale() == 'log':
        data = np.log10(data)
    xmin = quantile(data, q[0], weights)
    xmax = quantile(data, q[-1], weights)
    if 'range' in kwargs and ax.get_xaxis().get_scale() == 'log':
        range = kwargs.pop('range')
        if range is not None:
            range = (np.log10(range[0]), np.log10(range[1]))
        else:
            range = (data.min(), data.max())
    else:
        range = kwargs.pop('range', (xmin, xmax))
    if isinstance(bins, (int, str)):
        if isinstance(bins, int):
            bins = np.linspace(range[0], range[1], bins+1)
        elif isinstance(bins, str) and bins in ['fd', 'scott', 'sqrt']:
            bins = histogram_bin_edges(data,
                                       weights=weights,
                                       bins=bins,
                                       beta=kwargs.pop('beta', 'equal'),
                                       range=range)
        if ax.get_xaxis().get_scale() == 'log':
            bins = 10 ** bins
    if ax.get_xaxis().get_scale() == 'log':
        data = 10**data
        range = (10**range[0], 10**range[1])
    if isinstance(bins, str) and bins in ['knuth', 'freedman', 'blocks']:
        raise ValueError("The astropy strings 'knuth', 'freedman', and "
                         "'blocks' are no longer supported. Please use the"
                         "similar 'fd', 'scott', or 'sqrt' from now on.")
    else:
        h, edges, bars = ax.hist(data, weights=weights, bins=bins,
                                 range=range, histtype=histtype,
                                 color=color, *args, **kwargs)

    if histtype == 'bar' and not density:
        for b in bars:
            b.set_height(b.get_height() / h.max())
    elif (histtype == 'step' or histtype == 'stepfilled') and not density:
        trans = Affine2D().scale(sx=1, sy=1./h.max()) + ax.transData
        bars[0].set_transform(trans)

    if not density:
        ax.set_ylim(0, 1.1)
    return h, edges, bars


def fastkde_contour_plot_2d(ax, data_x, data_y, *args, **kwargs):
    """Plot a 2d marginalised distribution as contours.

    This functions as a wrapper around :meth:`matplotlib.axes.Axes.contour`,
    and :meth:`matplotlib.axes.Axes.contourf` with a kernel density estimation
    (KDE) computation in-between. All remaining keyword arguments are passed
    onwards to both functions.

    Parameters
    ----------
    ax : :class:`matplotlib.axes.Axes`
        Axis object to plot on.

    data_x, data_y : np.array
        The x and y coordinates of uniformly weighted samples to generate
        kernel density estimator.

    levels : list
        Amount of mass within each iso-probability contour.
        Has to be ordered from outermost to innermost contour.
        Default: [0.95, 0.68]

    xmin, xmax, ymin, ymax : float, default=None
        The lower/upper prior bounds in x/y coordinates.

    Returns
    -------
    c : :class:`matplotlib.contour.QuadContourSet`
        A set of contourlines or filled regions.

    """
    kwargs = normalize_kwargs(kwargs, dict(linewidths=['linewidth', 'lw'],
                                           linestyles=['linestyle', 'ls'],
                                           color=['c'],
                                           facecolor=['fc'],
                                           edgecolor=['ec']))

    xmin = kwargs.pop('xmin', None)
    xmax = kwargs.pop('xmax', None)
    ymin = kwargs.pop('ymin', None)
    ymax = kwargs.pop('ymax', None)
    if ax.get_xaxis().get_scale() == 'log':
        data_x = np.log10(data_x)
        xmin = None if xmin is None else np.log10(xmin)
        xmax = None if xmax is None else np.log10(xmax)
    if ax.get_yaxis().get_scale() == 'log':
        data_y = np.log10(data_y)
        ymin = None if ymin is None else np.log10(ymin)
        ymax = None if ymax is None else np.log10(ymax)
    label = kwargs.pop('label', None)
    zorder = kwargs.pop('zorder', 1)
    levels = kwargs.pop('levels', [0.95, 0.68])

    color = kwargs.pop('color', ax._get_lines.get_next_color())
    facecolor = kwargs.pop('facecolor', True)
    edgecolor = kwargs.pop('edgecolor', None)
    cmap = kwargs.pop('cmap', None)
    facecolor, edgecolor, cmap = set_colors(c=color, fc=facecolor,
                                            ec=edgecolor, cmap=cmap)

    kwargs.pop('q', None)

    try:
        from anesthetic.kde import fastkde_2d
        x, y, pdf, xmin, xmax, ymin, ymax = fastkde_2d(data_x, data_y,
                                                       xmin=xmin, xmax=xmax,
                                                       ymin=ymin, ymax=ymax)
    except ImportError:
        raise ImportError("You need to install fastkde to use fastkde")

    levels = iso_probability_contours(pdf, contours=levels)

    i = (pdf >= levels[0]*0.5).any(axis=0)
    j = (pdf >= levels[0]*0.5).any(axis=1)

    if ax.get_xaxis().get_scale() == 'log':
        x = 10**x
    if ax.get_yaxis().get_scale() == 'log':
        y = 10**y

    if facecolor not in [None, 'None', 'none']:
        linewidths = kwargs.pop('linewidths', 0.5)
        contf = ax.contourf(x[i], y[j], pdf[np.ix_(j, i)], levels, cmap=cmap,
                            zorder=zorder, vmin=0, vmax=pdf.max(),
                            *args, **kwargs)
        contf.set_cmap(cmap)
        ax.add_patch(plt.Rectangle((0, 0), 0, 0, lw=2, label=label,
                                   fc=cmap(0.999), ec=cmap(0.32)))
        cmap = None
    else:
        linewidths = kwargs.pop('linewidths',
                                plt.rcParams.get('lines.linewidth'))
        contf = None
        ax.add_patch(
            plt.Rectangle((0, 0), 0, 0, lw=2, label=label,
                          fc='None' if cmap is None else cmap(0.999),
                          ec=edgecolor if cmap is None else cmap(0.32))
        )

    vmin, vmax = match_contour_to_contourf(levels, vmin=0, vmax=pdf.max())
    cont = ax.contour(x[i], y[j], pdf[np.ix_(j, i)], levels, zorder=zorder,
                      vmin=vmin, vmax=vmax, linewidths=linewidths,
                      colors=edgecolor, cmap=cmap, *args, **kwargs)

    return contf, cont


def kde_contour_plot_2d(ax, data_x, data_y, *args, **kwargs):
    """Plot a 2d marginalised distribution as contours.

    This functions as a wrapper around :meth:`matplotlib.axes.Axes.contour`
    and :meth:`matplotlib.axes.Axes.contourf` with a kernel density
    estimation (KDE) computation provided by :class:`scipy.stats.gaussian_kde`
    in-between. All remaining keyword arguments are passed onwards to both
    functions.

    Parameters
    ----------
    ax : :class:`matplotlib.axes.Axes`
        Axis object to plot on.

    data_x, data_y : np.array
        The x and y coordinates of uniformly weighted samples to generate
        kernel density estimator.

    weights : np.array, optional
        Sample weights.

    levels : list, optional
        Amount of mass within each iso-probability contour.
        Has to be ordered from outermost to innermost contour.
        Default: [0.95, 0.68]

    ncompress : int, str, default='equal'
        Degree of compression.

        * If ``int``: desired number of samples after compression.
        * If ``False``: no compression.
        * If ``True``: compresses to the channel capacity, equivalent to
          ``ncompress='entropy'``.
        * If ``str``: determine number from the Huggins-Roy family of
          effective samples in :func:`anesthetic.utils.neff`
          with ``beta=ncompress``.

    nplot_2d : int, default=1000
        Number of plotting points to use.

    bw_method : str, scalar or callable, optional
        Forwarded to :class:`scipy.stats.gaussian_kde`.

    Returns
    -------
    c : :class:`matplotlib.contour.QuadContourSet`
        A set of contourlines or filled regions.

    """
    kwargs = normalize_kwargs(kwargs, dict(linewidths=['linewidth', 'lw'],
                                           linestyles=['linestyle', 'ls'],
                                           color=['c'],
                                           facecolor=['fc'],
                                           edgecolor=['ec']))

    weights = kwargs.pop('weights', None)
    if weights is not None:
        data_x = data_x[weights != 0]
        data_y = data_y[weights != 0]
        weights = weights[weights != 0]
    if ax.get_xaxis().get_scale() == 'log':
        data_x = np.log10(data_x)
    if ax.get_yaxis().get_scale() == 'log':
        data_y = np.log10(data_y)

    ncompress = kwargs.pop('ncompress', 'equal')
    nplot = kwargs.pop('nplot_2d', 1000)
    bw_method = kwargs.pop('bw_method', None)
    label = kwargs.pop('label', None)
    zorder = kwargs.pop('zorder', 1)
    levels = kwargs.pop('levels', [0.95, 0.68])

    color = kwargs.pop('color', ax._get_lines.get_next_color())
    facecolor = kwargs.pop('facecolor', True)
    edgecolor = kwargs.pop('edgecolor', None)
    cmap = kwargs.pop('cmap', None)
    facecolor, edgecolor, cmap = set_colors(c=color, fc=facecolor,
                                            ec=edgecolor, cmap=cmap)

    q = kwargs.pop('q', 5)
    q = quantile_plot_interval(q=q)
    xmin = quantile(data_x, q[0], weights)
    xmax = quantile(data_x, q[-1], weights)
    ymin = quantile(data_y, q[0], weights)
    ymax = quantile(data_y, q[-1], weights)
    X, Y = np.mgrid[xmin:xmax:1j*np.sqrt(nplot), ymin:ymax:1j*np.sqrt(nplot)]

    cov = np.cov(data_x, data_y, aweights=weights)
    tri, w = triangular_sample_compression_2d(data_x, data_y, cov,
                                              weights, ncompress)
    kde = gaussian_kde([tri.x, tri.y], weights=w, bw_method=bw_method)

    P = kde([X.ravel(), Y.ravel()]).reshape(X.shape)

    bw_x = np.sqrt(kde.covariance[0, 0])
    P = cut_and_normalise_gaussian(X, P, bw=bw_x,
                                   xmin=data_x.min(), xmax=data_x.max())
    bw_y = np.sqrt(kde.covariance[1, 1])
    P = cut_and_normalise_gaussian(Y, P, bw=bw_y,
                                   xmin=data_y.min(), xmax=data_y.max())

    levels = iso_probability_contours(P, contours=levels)
    if ax.get_xaxis().get_scale() == 'log':
        X = 10**X
    if ax.get_yaxis().get_scale() == 'log':
        Y = 10**Y

    if facecolor not in [None, 'None', 'none']:
        linewidths = kwargs.pop('linewidths', 0.5)
        contf = ax.contourf(X, Y, P, levels=levels, cmap=cmap, zorder=zorder,
                            vmin=0, vmax=P.max(), *args, **kwargs)
        contf.set_cmap(cmap)
        ax.add_patch(plt.Rectangle((0, 0), 0, 0, lw=2, label=label,
                                   fc=cmap(0.999), ec=cmap(0.32)))
        cmap = None
    else:
        linewidths = kwargs.pop('linewidths',
                                plt.rcParams.get('lines.linewidth'))
        contf = None
        ax.add_patch(
            plt.Rectangle((0, 0), 0, 0, lw=2, label=label,
                          fc='None' if cmap is None else cmap(0.999),
                          ec=edgecolor if cmap is None else cmap(0.32))
        )

    vmin, vmax = match_contour_to_contourf(levels, vmin=0, vmax=P.max())
    cont = ax.contour(X, Y, P, levels=levels, zorder=zorder,
                      vmin=vmin, vmax=vmax, linewidths=linewidths,
                      colors=edgecolor, cmap=cmap, *args, **kwargs)

    return contf, cont


def hist_plot_2d(ax, data_x, data_y, *args, **kwargs):
    """Plot a 2d marginalised distribution as a histogram.

    This functions as a wrapper around :meth:`matplotlib.axes.Axes.hist2d`.

    Parameters
    ----------
    ax : :class:`matplotlib.axes.Axes`
        Axis object to plot on.

    data_x, data_y : np.array
        The x and y coordinates of uniformly weighted samples to generate a
        two-dimensional histogram.

    levels : list, default=None
        Shade iso-probability contours containing these levels of probability
        mass. If None defaults to usual :meth:`matplotlib.axes.Axes.hist2d`
        colouring.

    q : int or float or tuple, default=5
        Quantile to determine the data range to be plotted.

        * ``0``: full data range, i.e. ``q=0`` --> quantile range (0, 1)
        * ``int``: q-sigma range, e.g. ``q=1`` --> quantile range (0.16, 0.84)
        * ``float``: percentile, e.g. ``q=0.8`` --> quantile range (0.1, 0.9)
        * ``tuple``: quantile range, e.g. (0.16, 0.84)

    Returns
    -------
    c : :class:`matplotlib.collections.QuadMesh`
        A set of colors.

    """
    kwargs = normalize_kwargs(kwargs)
    weights = kwargs.pop('weights', None)
    if ax.get_xaxis().get_scale() == 'log':
        data_x = np.log10(data_x)
    if ax.get_yaxis().get_scale() == 'log':
        data_y = np.log10(data_y)

    vmin = kwargs.pop('vmin', 0)
    label = kwargs.pop('label', None)
    levels = kwargs.pop('levels', None)

    color = kwargs.pop('color', ax._get_lines.get_next_color())
    cmap = kwargs.pop('cmap', basic_cmap(color))

    q = kwargs.pop('q', 5)
    q = quantile_plot_interval(q=q)
    xmin = quantile(data_x, q[0], weights)
    xmax = quantile(data_x, q[-1], weights)
    ymin = quantile(data_y, q[0], weights)
    ymax = quantile(data_y, q[-1], weights)
    rge = kwargs.pop('range', ((xmin, xmax), (ymin, ymax)))

    bins = kwargs.pop('bins', 10)
    density = kwargs.pop('density', False)
    cmin = kwargs.pop('cmin', None)
    cmax = kwargs.pop('cmax', None)
    pdf, x, y = np.histogram2d(data_x, data_y, bins, rge,
                               density, weights)
    if levels is not None:
        levels = iso_probability_contours(pdf, levels)
        pdf = np.digitize(pdf, levels, right=True)
        pdf = np.array(levels)[pdf]
        pdf = np.ma.masked_array(pdf, pdf < levels[1])
        if cmin is not None:
            pdf[pdf < cmin] = np.ma.masked
        if cmax is not None:
            pdf[pdf > cmax] = np.ma.masked
    snap = kwargs.pop('snap', True)
    if ax.get_xaxis().get_scale() == 'log':
        x = 10**x
    if ax.get_yaxis().get_scale() == 'log':
        y = 10**y
    image = ax.pcolormesh(x, y, pdf.T, cmap=cmap, vmin=vmin, snap=snap,
                          *args, **kwargs)

    ax.add_patch(plt.Rectangle((0, 0), 0, 0, fc=cmap(0.999), ec=cmap(0.32),
                               lw=2, label=label))

    return image


def scatter_plot_2d(ax, data_x, data_y, *args, **kwargs):
    """Plot samples from a 2d marginalised distribution.

    This functions as a wrapper around :meth:`matplotlib.axes.Axes.plot`,
    enforcing any prior bounds. All remaining keyword arguments are passed
    onwards.

    Parameters
    ----------
    ax : :class:`matplotlib.axes.Axes`
        axis object to plot on

    data_x, data_y : np.array
        x and y coordinates of uniformly weighted samples to plot.

    ncompress : int, str, default='equal'
        Degree of compression.

        * If ``int``: desired number of samples after compression.
        * If ``False``: no compression.
        * If ``True``: compresses to the channel capacity, equivalent to
          ``ncompress='entropy'``.
        * If ``str``: determine number from the Huggins-Roy family of
          effective samples in :func:`anesthetic.utils.neff`
          with ``beta=ncompress``.

    Returns
    -------
    lines : :class:`matplotlib.lines.Line2D`
        A list of line objects representing the plotted data (same as
        :meth:`matplotlib.axes.Axes.plot` command).

    """
    kwargs = normalize_kwargs(
        kwargs,
        alias_mapping=dict(lw=['linewidth', 'linewidths'],
                           ls=['linestyle', 'linestyles'],
                           color=['c'],
                           mfc=['fc', 'facecolor'],
                           mec=['ec', 'edgecolor'],
                           cmap=['colormap']),
        drop=['ls', 'lw']
    )
    kwargs = cbook.normalize_kwargs(kwargs, mlines.Line2D)

    markersize = kwargs.pop('markersize', 1)
    cmap = kwargs.pop('cmap', None)
    color = kwargs.pop('color', (ax._get_lines.get_next_color()
                                 if cmap is None else cmap(0.68)))

    weights = kwargs.pop('weights', None)
    q = kwargs.pop('q', 5)
    q = quantile_plot_interval(q=q)
    xmin = quantile(data_x, q[0], weights)
    xmax = quantile(data_x, q[-1], weights)
    ymin = quantile(data_y, q[0], weights)
    ymax = quantile(data_y, q[-1], weights)
    mask = ((data_x >= xmin) & (data_x <= xmax) &
            (data_y >= ymin) & (data_y <= ymax))

    points = ax.plot(data_x[mask], data_y[mask], 'o',
                     color=color, markersize=markersize, *args, **kwargs)
    return points


def basic_cmap(color):
    """Construct basic colormap a single color."""
    return LinearSegmentedColormap.from_list(str(color), ['#ffffff', color])


def quantile_plot_interval(q):
    """Interpret quantile ``q`` input to quantile plot range tuple."""
    if isinstance(q, str):
        sigmas = {'1sigma': 0.682689492137086,
                  '2sigma': 0.954499736103642,
                  '3sigma': 0.997300203936740,
                  '4sigma': 0.999936657516334,
                  '5sigma': 0.999999426696856}
        q = (1 - sigmas[q]) / 2
    elif isinstance(q, int) and q >= 1:
        q = (1 - erf(q / np.sqrt(2))) / 2
    if isinstance(q, float) or isinstance(q, int):
        if q > 0.5:
            q = 1 - q
        q = (q, 1-q)
    return tuple(np.sort(q))


def normalize_kwargs(kwargs, alias_mapping=None, drop=None):
    """Normalize kwarg inputs.

    Works the same way as :func:`matplotlib.cbook.normalize_kwargs`, but
    additionally allows to drop kwargs.
    """
    drop = [] if drop is None else drop
    if alias_mapping is None:
        alias_mapping = dict(linewidth=['lw'],
                             linestyle=['ls'],
                             color=['c'],
                             facecolor=['fc'],
                             edgecolor=['ec'],
                             cmap=['colormap'])
    kwargs = cbook.normalize_kwargs(kwargs, alias_mapping=alias_mapping)
    for key in set(drop) & set(kwargs.keys()):
        kwargs.pop(key)
    return kwargs


def set_colors(c, fc, ec, cmap):
    """Navigate interplay between possible color inputs {c, fc, ec, cmap}."""
    if fc in [None, 'None', 'none']:
        # unfilled contours
        if ec is None and cmap is None:
            cmap = basic_cmap(c)
    else:
        # filled contours
        if fc is True:
            fc = c
        if ec is None and cmap is None:
            ec = c
            cmap = basic_cmap(fc)
        elif ec is None:
            ec = (cmap(1.),)
        elif cmap is None:
            cmap = basic_cmap(fc)
    return fc, ec, cmap



import matplotlib.pyplot as plt
from anesthetic.scripts import gui
from bin.utils import unit_incremented
import pytest


@pytest.fixture(autouse=True)
def close_figures_on_teardown():
    yield
    plt.close("all")


def test_gui():
    gui(['./tests/example_data/pc', '--params', 'x0', 'x1', 'x2'])


@pytest.mark.parametrize('a, b', [['2.0.0b2', '2.0.0b1'],
                                  ['2.0.0b0', '2.0.0a3'],
                                  ['2.0.0', '2.0.0b1'],
                                  ['3.0.0a0', '2.5.6'],
                                  ['2.0.3', '2.0.2'],
                                  ['2.1.0', '2.0.5'],
                                  ['3.0.0', '2.5.6'],
                                  ])
def test_is_unit_incremented(a, b):
    assert unit_incremented(a, b)


@pytest.mark.parametrize('a, b', [['2.0.0b3', '2.0.0b1'],
                                  ['2.0.0b3', '2.0.0a3'],
                                  ['2.0.1', '2.0.0b1'],
                                  ['3.0.0a1', '2.5.6'],
                                  ['2.0.4', '2.0.2'],
                                  ['2.1.5', '2.0.5'],
                                  ['3.5.6', '2.5.6'],
                                  ])
def test_is_not_unit_incremented(a, b):
    assert not unit_incremented(a, b)



from __future__ import annotations
from matplotlib import pyplot as plt
import numpy as np
from pandas.plotting._matplotlib.hist import (HistPlot as _HistPlot,
                                              KdePlot as _KdePlot)
import pandas.plotting._matplotlib.hist
from pandas.core.dtypes.missing import (
    isna,
    remove_na_arraylike,
)
from pandas.plotting._matplotlib.core import MPLPlot
from typing import Literal
from anesthetic.plotting._matplotlib.core import (
    _WeightedMPLPlot, _CompressedMPLPlot, _PlanePlot2d, _get_weights
)
from anesthetic.plot import (
    kde_contour_plot_2d,
    hist_plot_2d,
    fastkde_contour_plot_2d,
    kde_plot_1d,
    fastkde_plot_1d,
    hist_plot_1d,
    quantile_plot_interval,
)
from anesthetic.utils import histogram_bin_edges


class HistPlot(_WeightedMPLPlot, _HistPlot):

    # noqa: disable=D101
    def _adjust_bins(self, bins) -> None:
        if (
                isinstance(bins, str) and
                bins in ['fd', 'scott', 'sqrt']
        ):
            bins = self._calculate_bins(self.data, bins)
        return super()._adjust_bins(bins)

    # noqa: disable=D101
    def _calculate_bins(self, data, bins):
        if self.logx:
            data = np.log10(data)
            if self._bin_range is not None:
                xmin, xmax = self._bin_range
                self._bin_range = (np.log10(xmin), np.log10(xmax))
        nd_values = data.infer_objects()._get_numeric_data()
        values = np.ravel(nd_values)
        weights = self.weights
        if weights is not None:
            weights = np.broadcast_to(weights[:, None], nd_values.shape)
            weights = np.ravel(weights)
            weights = weights[~isna(values)]

        values = values[~isna(values)]

        if isinstance(bins, str) and bins in ['fd', 'scott', 'sqrt']:
            bins = histogram_bin_edges(
                values,
                weights=weights,
                bins=bins,
                beta=self.kwds.pop('beta', 'equal'),
                range=self._bin_range
            )
        else:
            bins = np.histogram_bin_edges(
                values,
                weights=weights,
                bins=bins,
                range=self._bin_range
            )
        if self.logx:
            bins = 10**bins
            if self._bin_range is not None:
                self._bin_range = (xmin, xmax)
        return bins

    def _get_colors(self, num_colors=None, color_kwds='color'):
        if (self.colormap is not None and self.kwds.get(color_kwds) is None
           and (num_colors is None or num_colors == 1)):
            return [plt.get_cmap(self.colormap)(0.68)]
        return super()._get_colors(num_colors, color_kwds)

    def _post_plot_logic(self, ax, data):
        ax.set_xlabel(self.xlabel)
        ax.set_yticks([])
        ax.set_ylim(bottom=0)


class KdePlot(HistPlot, _KdePlot):
    # noqa: disable=D101
    @classmethod
    def _plot(
        cls,
        ax,
        y,
        style=None,
        bw_method=None,
        ind=None,
        column_num=None,
        stacking_id=None,
        **kwds,
    ):
        from scipy.stats import gaussian_kde
        weights = kwds.pop("weights", None)

        y = remove_na_arraylike(y)
        gkde = gaussian_kde(y, bw_method=bw_method, weights=weights)

        y = gkde.evaluate(ind)
        lines = MPLPlot._plot(ax, ind, y, style=style, **kwds)
        return lines

    def _post_plot_logic(self, ax, data):
        ax.set_yticks([])
        ax.set_ylim(bottom=0)


class Kde1dPlot(KdePlot):
    # noqa: disable=D101
    @property
    def _kind(self) -> Literal["kde_1d"]:
        return "kde_1d"

    # noqa: disable=D101
    @classmethod
    def _plot(
        cls,
        ax,
        y,
        style=None,
        ind=None,
        column_num=None,
        stacking_id=None,
        **kwds,
    ):
        args = (style,) if style is not None else tuple()
        return kde_plot_1d(ax, y, *args, **kwds)


class FastKde1dPlot(_CompressedMPLPlot, Kde1dPlot):
    # noqa: disable=D101
    @property
    def _kind(self) -> Literal["fastkde_1d"]:
        return "fastkde_1d"

    @classmethod
    def _plot(
        cls,
        ax,
        y,
        style=None,
        ind=None,
        column_num=None,
        stacking_id=None,
        **kwds,
    ):
        args = (style,) if style is not None else tuple()
        # weights and bw_method are not valid for fastkde
        kwds.pop('weights', None)
        kwds.pop('bw_method', None)
        return fastkde_plot_1d(ax, y, *args, **kwds)


class Hist1dPlot(HistPlot):
    # noqa: disable=D101
    @property
    def _kind(self) -> Literal["hist_1d"]:
        return "hist_1d"

    def __init__(
            self,
            data,
            bins: str | int | np.ndarray | list[np.ndarray] = 'fd',
            bottom: int | np.ndarray = 0,
            **kwargs,
    ) -> None:
        super().__init__(data, bins=bins, bottom=bottom, **kwargs)

    def _calculate_bins(self, data, bins):
        if self._bin_range is None:
            q = self.kwds.get('q', 5)
            q = quantile_plot_interval(q=q)
            xmin, xmax = data.quantile(q).to_numpy().ravel()
            self._bin_range = (xmin, xmax)
        bins = super()._calculate_bins(data, bins)
        return bins

    @classmethod
    def _plot(
        cls,
        ax,
        y,
        style=None,
        bottom=0,
        column_num=None,
        stacking_id=None,
        *,
        bins,
        **kwds,
    ):
        if column_num == 0:
            cls._initialize_stacker(ax, stacking_id, len(bins) - 1)

        if not isinstance(bins, str):
            base = np.zeros(len(bins) - 1)
            bottom = bottom + cls._get_stacked_values(ax, stacking_id,
                                                      base, kwds["label"])

        # ignore style
        n, bins, patches = hist_plot_1d(ax, y, bins=bins,
                                        bottom=bottom, **kwds)
        cls._update_stacker(ax, stacking_id, n)
        return patches


class Kde2dPlot(_WeightedMPLPlot, _PlanePlot2d):
    # noqa: disable=D101
    @property
    def _kind(self) -> Literal["kde_2d"]:
        return "kde_2d"

    @classmethod
    def _plot(cls, ax, x, y, **kwds):
        return kde_contour_plot_2d(ax, x, y, **kwds)


class FastKde2dPlot(_CompressedMPLPlot, _PlanePlot2d):
    # noqa: disable=D101
    @property
    def _kind(self) -> Literal["fastkde_2d"]:
        return "fastkde_2d"

    @classmethod
    def _plot(cls, ax, x, y, **kwds):
        return fastkde_contour_plot_2d(ax, x, y, **kwds)


class Hist2dPlot(_WeightedMPLPlot, _PlanePlot2d):
    # noqa: disable=D101
    @property
    def _kind(self) -> Literal["hist_2d"]:
        return "hist_2d"

    @classmethod
    def _plot(cls, ax, x, y, **kwds):
        return hist_plot_2d(ax, x, y, **kwds)


def hist_frame(data, *args, **kwds):
    # noqa: disable=D103
    _get_weights(kwds, data)
    return pandas.plotting._matplotlib.hist_frame(data, *args, **kwds)


def hist_series(data, *args, **kwds):
    # noqa: disable=D103
    _get_weights(kwds, data)
    return pandas.plotting._matplotlib.hist_series(data, *args, **kwds)



from anesthetic import read_chains
from anesthetic.convert import to_getdist
from numpy.testing import assert_array_equal
from utils import getdist_mark_xfail


@getdist_mark_xfail
def test_to_getdist():
    anesthetic_samples = read_chains('./tests/example_data/gd')
    getdist_samples = to_getdist(anesthetic_samples)

    assert_array_equal(getdist_samples.samples, anesthetic_samples)
    assert_array_equal(getdist_samples.weights,
                       anesthetic_samples.get_weights())

    for param, p in zip(getdist_samples.getParamNames().names,
                        anesthetic_samples.drop_labels().columns):

        assert param.name == p



import matplotlib.pyplot as plt
import anesthetic.examples._matplotlib_agg  # noqa: F401
from anesthetic import read_chains
import pytest
from pytest import approx
from utils import skipif_no_h5py


@pytest.fixture(autouse=True)
def close_figures_on_teardown():
    yield
    plt.close("all")


@pytest.mark.parametrize('root', ["./tests/example_data/pc",
                                  "./tests/example_data/mn",
                                  skipif_no_h5py("./tests/example_data/un"),
                                  "./tests/example_data/nf"])
def test_gui(root):
    samples = read_chains(root)
    plotter = samples.gui()

    # Type buttons
    plotter.type.buttons.set_active(1)
    assert plotter.type() == 'posterior'
    plotter.type.buttons.set_active(0)
    assert plotter.type() == 'live'

    # Parameter choice buttons
    plotter.param_choice.buttons.set_active(1)
    assert len(plotter.triangle.ax) == 2
    plotter.param_choice.buttons.set_active(0)
    assert len(plotter.triangle.ax) == 1
    plotter.param_choice.buttons.set_active(0)
    plotter.param_choice.buttons.set_active(2)
    plotter.param_choice.buttons.set_active(3)
    assert len(plotter.triangle.ax) == 4

    # Sliders
    old = plotter.evolution()
    plotter.evolution.slider.set_val(5)
    assert plotter.evolution() != old
    plotter.evolution.slider.set_val(0)
    assert plotter.evolution() == old
    plotter.type.buttons.set_active(1)

    plotter.beta.slider.set_val(0)
    assert plotter.beta() == approx(0, rel=0, abs=1e-8)

    plotter.beta.slider.set_val(samples.D_KL())
    assert plotter.beta() == approx(1)
    plotter.beta.slider.set_val(1e2)
    assert plotter.beta() == approx(min(plotter.samples.beta_max(), 1e10))
    plotter.type.buttons.set_active(0)

    # Reload button
    plotter.reload.button.on_clicked(plotter.reload_file(None))

    # Reset button
    plotter.reset.button.on_clicked(plotter.reset_range(None))


@pytest.mark.parametrize('root', ["./tests/example_data/pc",
                                  "./tests/example_data/mn",
                                  skipif_no_h5py("./tests/example_data/un"),
                                  "./tests/example_data/nf"])
def test_gui_params(root):
    samples = read_chains(root)
    params = samples.columns.get_level_values(0).to_list()
    plotter = samples.gui()
    assert len(plotter.param_choice.buttons.labels) == len(params)

    plotter = samples.gui(params=params[:2])
    assert len(plotter.param_choice.buttons.labels) == 2


@pytest.mark.parametrize('root', ["./tests/example_data/pc",
                                  "./tests/example_data/mn",
                                  skipif_no_h5py("./tests/example_data/un"),
                                  "./tests/example_data/nf"])
def test_slider_reset_range(root):
    plotter = read_chains(root).gui()
    plotter.evolution.reset_range(-3, 3)
    assert plotter.evolution.ax.get_xlim() == (-3.0, 3.0)



"""Tests for WeightedSeries/DataFrame consistency with pandas behavior."""
import warnings
import numpy as np
import pandas as pd
import pytest
from anesthetic.weighted_pandas import WeightedSeries, WeightedDataFrame


class TestPandasConsistency:
    """Test consistency with pandas for unweighted data."""

    @pytest.mark.parametrize('data', [[np.nan, np.nan, np.nan, np.nan, np.nan],
                                      [1.0, np.nan, 3.0, 4.0, 5.0, 9.0],
                                      [2.0],
                                      [2.0, 3.0],
                                      [2.0, 3.0, 9.0],
                                      [2.0, 3.0, 9.0, 2.0],
                                      [np.nan],
                                      [0.0, np.nan, 0.0, 0.0, 0.0],
                                      [1e-4, 2e-4, 3e-4, 4e-4, 9e-4],
                                      [1e10, 2e10, 3e10, 4e10, 9e10]])
    @pytest.mark.parametrize('method, ddof', [
        ('mean', None),
        ('var', 0),
        ('var', 1),
        ('std', 0),
        ('std', 1),
        ('sem', 0),
        ('sem', 1),
        ('skew', None),
        ('kurt', None),
        ('cov', None),
        ('corr', None),
    ])
    @pytest.mark.parametrize('skipna', [True, False])
    @pytest.mark.parametrize('weight_type', ['frequency', 'reliability'])
    def test_series_consistency(self, data, method, ddof, skipna, weight_type):
        """Test WeightedSeries consistency with pandas Series."""
        if method in ['cov', 'corr'] and (np.isfinite(data).sum() <= 1 or
                                          np.nanvar(data) == 0):
            warnings.filterwarnings(action='ignore', category=RuntimeWarning)
        else:
            warnings.resetwarnings()
        ps = pd.Series(data)
        if weight_type == 'frequency':
            weights = np.ones_like(data, dtype=int)
            ps = ps.loc[ps.index.repeat(weights)].reset_index(drop=True)
        else:
            weights = np.ones_like(data, dtype=float) / len(data)
        ws = WeightedSeries(data, weights=weights)  # equally weighted
        ds = ws.drop_weights()
        assert ws.isweighted() is True
        assert ds.isweighted() is False

        # Get results
        if ddof is not None:
            pandas_result = getattr(ps, method)(skipna=skipna, ddof=ddof)
            weight_result = getattr(ws, method)(skipna=skipna, ddof=ddof)
            droppd_result = getattr(ds, method)(skipna=skipna, ddof=ddof)
        elif method in ['cov', 'corr']:
            pandas_result = getattr(ps, method)(ps)
            weight_result = getattr(ws, method)(ws)
            droppd_result = getattr(ds, method)(ds)
        else:
            pandas_result = getattr(ps, method)(skipna=skipna)
            weight_result = getattr(ws, method)(skipna=skipna)
            droppd_result = getattr(ds, method)(skipna=skipna)

        # Check consistency
        assert np.isnan(weight_result) is np.isnan(pandas_result)
        assert np.isnan(droppd_result) is np.isnan(pandas_result)
        assert weight_result == pytest.approx(pandas_result, rel=1e-12,
                                              nan_ok=True)
        assert droppd_result == pytest.approx(pandas_result, rel=1e-12,
                                              nan_ok=True)
        assert type(weight_result) is type(pandas_result)
        assert type(droppd_result) is type(pandas_result)

    @pytest.mark.parametrize('data', [
        np.array([[1.0, 2.0, 3.0],
                  [4.0, 5.0, 6.0],
                  [7.0, 8.0, 9.0],
                  [7.0, 8.0, 9.0]]),
        np.array([[1.0, 2.0],
                  [3.0, 4.0],
                  [7.0, 8.0]]),
        np.array([[1.0, 2.0],
                  [7.0, 8.0]]),
        np.array([[1.0, 2.0]]),
        np.array([[1.0],
                  [4.0],
                  [7.0],
                  [7.0]]),
        np.array([[1.0]]),
        np.array([[np.nan]]),
        np.array([[1.0, 2.0, np.nan],
                  [4.0, np.nan, 6.0],
                  [7.0, 8.0, np.nan],
                  [7.0, 8.0, 9.0]]),
        np.array([[np.nan, np.nan, np.nan],
                  [np.nan, np.nan, np.nan],
                  [np.nan, np.nan, np.nan],
                  [np.nan, np.nan, np.nan]]),
    ])
    @pytest.mark.parametrize('method, ddof, skipna', [
        ('mean', None, True),
        ('mean', None, False),
        ('var', 0, True), ('var', 0, False),
        ('var', 1, True), ('var', 1, False),
        ('std', 0, True), ('std', 0, False),
        ('std', 1, True), ('std', 1, False),
        ('sem', 0, True), ('sem', 0, False),
        ('sem', 1, True), ('sem', 1, False),
        # ('cov', 0, None),  # not testing because of pandas bug: issue #45814
        ('cov', 1, None),
        ('corr', None, None),
        ('skew', None, True), ('skew', None, False),
        ('kurt', None, True), ('kurt', None, False),
    ])
    @pytest.mark.parametrize('axis', [0, 1])
    @pytest.mark.parametrize('weight_type', ['frequency', 'reliability'])
    def test_dataframe_consistency(self, data, method, ddof, skipna, axis,
                                   weight_type):
        """Test WeightedDataFrame consistency with pandas DataFrame."""
        if method in ['cov', 'corr'] and axis == 1:
            pytest.skip(f"`{method}` does not support `axis` kwarg")
        elif data.shape[0] == 1:
            warnings.filterwarnings(action='ignore', category=RuntimeWarning)
        else:
            warnings.resetwarnings()
        columns = ['A', 'B', 'C'][:data.shape[1]]
        pdf = pd.DataFrame(data, columns=columns)
        if weight_type == 'frequency':
            weights = np.ones(data.shape[0], dtype=int) * 3
            if axis == 0:
                pdf = pdf.loc[pdf.index.repeat(weights)].reset_index(drop=True)
        else:
            weights = np.ones(data.shape[0], dtype=float) / data.shape[0]
        wdf = WeightedDataFrame(data, columns=columns, weights=weights)

        # Get results
        if skipna is None and ddof is None:
            # for corr
            pandas_result = getattr(pdf, method)()
            weight_result = getattr(wdf, method)()
        elif skipna is None:
            # for cov
            pandas_result = getattr(pdf, method)(ddof=ddof)
            with pytest.raises(TypeError):
                getattr(wdf, method)(ddof=ddof, skipna=True, axis=1)
            weight_result = getattr(wdf, method)(ddof=ddof)
        elif ddof is not None:
            # for var, std, sem
            pandas_result = getattr(pdf, method)(skipna=skipna, axis=axis,
                                                 ddof=ddof)
            weight_result = getattr(wdf, method)(skipna=skipna, axis=axis,
                                                 ddof=ddof)
        else:
            # for mean, skew, kurt
            pandas_result = getattr(pdf, method)(skipna=skipna, axis=axis)
            weight_result = getattr(wdf, method)(skipna=skipna, axis=axis)

        # Check consistency
        pandas_vals = pandas_result.values
        weight_vals = weight_result.values

        assert pandas_vals.shape == weight_vals.shape
        both_nan = np.isnan(pandas_vals) & np.isnan(weight_vals)
        values_match = np.allclose(pandas_vals, weight_vals,
                                   equal_nan=True, rtol=1e-10)

        assert np.all(both_nan | values_match), (
            f"{method}(skipna={skipna}, axis={axis}, ddof={ddof}) "
            f"inconsistent for {weight_type} weights: \ndata =\n{data}, \n"
            f"pandas =\n{pandas_vals}, \nweighted =\n{weight_vals}"
        )
        assert type(weight_vals) is type(pandas_vals)

    def test_weighted_data_differs_from_pandas(self):
        """Test that truly weighted data gives different results."""
        data = [1.0, 2.0, 3.0, 4.0, 9.0]
        weights = [0.1, 0.2, 0.3, 0.3, 0.1]  # Non-uniform weights

        ps = pd.Series(data)
        ws = WeightedSeries(data, weights=weights)

        # These should be different (weighted vs unweighted)
        assert not np.allclose(ps.mean(), ws.mean())
        assert not np.allclose(ps.var(ddof=0), ws.var(ddof=0))
        assert not np.allclose(ps.var(ddof=1), ws.var(ddof=1))
        assert not np.allclose(ps.std(ddof=0), ws.std(ddof=0))
        assert not np.allclose(ps.std(ddof=1), ws.std(ddof=1))
        assert not np.allclose(ps.skew(), ws.skew())
        assert not np.allclose(ps.kurt(), ws.kurt())

    def test_edge_cases(self):
        """Test edge cases that should be handled consistently."""
        # Empty data
        empty_ps = pd.Series([], dtype=float)
        empty_ws = WeightedSeries([], dtype=float)

        assert np.isnan(empty_ps.mean()) == np.isnan(empty_ws.mean())
        assert np.isnan(empty_ps.var()) == np.isnan(empty_ws.var())

        # All zero weights should return NaN
        data = [1.0, 2.0, 3.0]
        ws_zero_weights = WeightedSeries(data, weights=[0.0, 0.0, 0.0])

        assert np.isnan(ws_zero_weights.mean())
        assert np.isnan(ws_zero_weights.var())
        assert np.isnan(ws_zero_weights.std())

    def test_mathematical_properties(self):
        """Test that mathematical properties are preserved."""
        data = [1.0, 2.0, 3.0, 4.0, 5.0]
        ws = WeightedSeries(data)  # Unweighted

        # std should be sqrt of var
        assert np.allclose(ws.std(), np.sqrt(ws.var()))

        # Mean of constants should be the constant
        constant_ws = WeightedSeries([5.0, 5.0, 5.0, 5.0])
        assert np.allclose(constant_ws.mean(), 5.0)
        assert np.allclose(constant_ws.var(), 0.0)

        # Variance of constants should be zero
        assert np.allclose(constant_ws.var(), 0.0)



import anesthetic.examples._matplotlib_agg  # noqa: F401
from packaging import version
from warnings import catch_warnings, filterwarnings
import pytest
import numpy as np
import matplotlib
import matplotlib.pyplot as plt
import matplotlib.gridspec as gs
from anesthetic.plot import (make_1d_axes, make_2d_axes, kde_plot_1d,
                             fastkde_plot_1d, hist_plot_1d, hist_plot_2d,
                             fastkde_contour_plot_2d, kde_contour_plot_2d,
                             scatter_plot_2d, quantile_plot_interval,
                             basic_cmap, AxesSeries, AxesDataFrame)
from numpy.testing import assert_array_equal

from matplotlib.axes import SubplotBase
from matplotlib.contour import ContourSet
from matplotlib.lines import Line2D
from matplotlib.patches import Patch, Polygon
from matplotlib.colors import ColorConverter, to_rgba
from matplotlib.figure import Figure
from scipy.special import erf
from scipy.interpolate import interp1d
from scipy import stats
from utils import skipif_no_fastkde, astropy_mark_xfail, fastkde_mark_xfail


@pytest.fixture(autouse=True)
def close_figures_on_teardown():
    yield
    plt.close("all")


def test_AxesObjects():
    paramnames = ['a', 'b', 'c']

    # AxesSeries
    axes = AxesSeries(index=paramnames)
    assert isinstance(axes, AxesSeries)
    assert isinstance(axes.iloc[0], SubplotBase)
    assert axes.iloc[0].get_xlabel() == 'a'
    axes.set_xlabels(labels=dict(a='A', b='B', c='C'))
    assert axes.iloc[0].get_xlabel() == 'A'
    axes.tick_params(labelrotation=0, labelsize='medium')

    # AxesDataFrame
    axes = AxesDataFrame(index=paramnames + ['d'], columns=paramnames)
    assert isinstance(axes, AxesDataFrame)
    assert isinstance(axes.iloc[0, 0], SubplotBase)
    assert axes.iloc[-1, 0].get_xlabel() == 'a'
    assert axes.iloc[-1, 0].get_ylabel() == 'd'
    axes.set_labels(labels=dict(a='A', b='B', c='C', d='D'))
    assert axes.iloc[-1, 0].get_xlabel() == 'A'
    assert axes.iloc[-1, 0].get_ylabel() == 'D'
    axes.tick_params(labelrotation=0, labelsize='medium')
    xmin1, xmax1 = axes.iloc[0, 0].get_xlim()
    ymin1, ymax1 = axes.iloc[0, 0].get_ylim()
    axes.set_margins(m=0.5)
    xmin2, xmax2 = axes.iloc[0, 0].get_xlim()
    ymin2, ymax2 = axes.iloc[0, 0].get_ylim()
    assert xmin2 == xmin1 - 0.5 * (xmax1 - xmin1)
    assert xmax2 == xmax1 + 0.5 * (xmax1 - xmin1)
    assert ymin2 == ymin1 - 0.5 * (ymax1 - ymin1)
    assert ymax2 == ymax1 + 0.5 * (ymax1 - ymin1)


def test_make_1d_axes():
    paramnames = ['A', 'B', 'C', 'D', 'E']
    labels = {'A': 'tA', 'B': 'tB', 'C': 'tC', 'D': 'tD', 'E': 'tE'}

    # Check no optional arguments
    fig, axes = make_1d_axes(paramnames)
    assert isinstance(fig, Figure)
    assert isinstance(axes, AxesSeries)
    assert isinstance(axes.to_frame(), AxesDataFrame)
    assert_array_equal(axes.index, paramnames)
    for p, ax in axes.items():
        assert ax.get_xlabel() == p

    # Check single string input
    fig, axes = make_1d_axes(paramnames[0])
    assert isinstance(fig, Figure)
    assert isinstance(axes, AxesSeries)
    assert axes.index.size == 1
    assert_array_equal(axes.index, paramnames[0])

    # Check labels argument
    fig, axes = make_1d_axes(paramnames, labels=labels)
    for t in labels:
        assert axes[t].get_xlabel() != t
        assert axes[t].get_xlabel() == labels[t]

    # Check fig argument
    fig0 = plt.figure()
    fig0.suptitle('hi there')
    fig, axes = make_1d_axes(paramnames)
    assert fig is not fig0
    fig, axes = make_1d_axes(paramnames, fig=fig0)
    assert fig is fig0

    # Check ncol argument
    fig, axes = make_1d_axes(paramnames, ncol=2)
    nrows, ncol = axes.iloc[0].get_subplotspec().get_gridspec().get_geometry()
    assert ncol == 2

    # Check gridspec argument
    grid = gs.GridSpec(2, 2, width_ratios=[3, 1], height_ratios=[3, 1])
    g00 = grid[0, 0]
    fig, axes = make_1d_axes(paramnames, subplot_spec=g00)
    assert g00 is axes.iloc[0].get_subplotspec().get_topmost_subplotspec()

    # Check gridspec kwargs
    fig, axes = make_1d_axes(paramnames, gridspec_kw=dict(wspace=0.1))
    ws = axes['A'].get_subplotspec().get_gridspec().get_subplot_params().wspace
    assert ws == 0.1

    # Check figure kwargs
    fig, axes = make_1d_axes(paramnames, figsize=(5, 5))
    assert fig.get_figwidth() == 5
    assert fig.get_figheight() == 5

    # Check unexpected kwargs
    with pytest.raises((AttributeError, TypeError)):
        make_1d_axes(paramnames, spam='ham')


@pytest.mark.parametrize('params', [[0, 1, 2, 3],
                                    [0, 2, 3],
                                    [1, 2, 3],
                                    ['x0', 'x1', 'x2'],
                                    ['1', '2', '3'],
                                    [0, 'x1', 'x2'],
                                    ])
def test_make_Nd_axes_integers(params):
    fig, axes = make_1d_axes(params)
    assert isinstance(fig, Figure)
    assert isinstance(axes, AxesSeries)
    assert list(axes.index) == params

    fig, axes = make_2d_axes(params)
    assert isinstance(fig, Figure)
    assert isinstance(axes, AxesDataFrame)
    assert list(axes.index) == params
    assert list(axes.columns) == params


def test_make_2d_axes_inputs_outputs():
    paramnames_x = ['A', 'B', 'C', 'D']
    paramnames_y = ['B', 'A', 'D', 'E']

    # 2D axes
    fig, axes = make_2d_axes([paramnames_x, paramnames_y])
    assert isinstance(fig, Figure)
    assert isinstance(axes, AxesDataFrame)
    assert isinstance(axes['A'], AxesSeries)
    assert isinstance(axes.loc['A':'B', 'B':'C'], AxesDataFrame)
    assert_array_equal(axes.index, paramnames_y)
    assert_array_equal(axes.columns, paramnames_x)

    # Axes labels
    for p, ax in axes.iloc[:, 0].items():
        assert ax.get_ylabel() == p

    for p, ax in axes.iloc[-1].items():
        assert ax.get_xlabel() == p

    for ax in axes.iloc[:-1, 1:].to_numpy().flatten():
        assert ax.get_xlabel() == ''
        assert ax.get_ylabel() == ''

    # Check fig argument
    fig0 = plt.figure()
    fig, axes = make_2d_axes(paramnames_x)
    assert fig is not fig0
    fig, axes = make_2d_axes(paramnames_x, fig=fig0)
    assert fig is fig0

    # Check gridspec argument
    grid = gs.GridSpec(2, 2, width_ratios=[3, 1], height_ratios=[3, 1])
    g00 = grid[0, 0]
    fig, axes = make_2d_axes(paramnames_x, subplot_spec=g00)
    assert g00 is axes.iloc[0, 0].get_subplotspec().get_topmost_subplotspec()

    # Check gridspec kwargs
    fig, axes = make_1d_axes(paramnames_x, gridspec_kw=dict(wspace=0.1))
    ws = axes['A'].get_subplotspec().get_gridspec().get_subplot_params().wspace
    assert ws == 0.1

    # Check figure kwargs
    fig, axes = make_2d_axes(paramnames_x, figsize=(5, 5))
    assert fig.get_figwidth() == 5
    assert fig.get_figheight() == 5

    # Check unexpected kwargs
    with pytest.raises((AttributeError, TypeError)):
        make_2d_axes(paramnames_x, spam='ham')


@pytest.mark.parametrize('paramnames_y', [['A', 'B', 'C', 'D'],
                                          ['A', 'C', 'B', 'D'],
                                          ['D', 'C', 'B', 'A'],
                                          ['C', 'B', 'A'],
                                          ['E', 'F', 'G', 'H'],
                                          ['A', 'B', 'E', 'F'],
                                          ['B', 'E', 'A', 'F'],
                                          ['B', 'F', 'A', 'H', 'G'],
                                          ['B', 'A', 'H', 'G']])
@pytest.mark.parametrize('upper', [False, True])
@pytest.mark.parametrize('lower', [False, True])
@pytest.mark.parametrize('diagonal', [False, True])
def test_make_2d_axes_behaviour(diagonal, lower, upper, paramnames_y):
    np.random.seed(0)

    def calc_n(axes):
        """Compute the number of upper, lower and diagonal plots."""
        n = {'upper': 0, 'lower': 0, 'diagonal': 0}
        for y, row in axes.iterrows():
            for x, ax in row.items():
                if ax is not None:
                    n[ax.position] += 1
        return n

    # Check for only paramnames_x
    paramnames_x = ['A', 'B', 'C', 'D']
    nx = len(paramnames_x)
    fig, axes = make_2d_axes(paramnames_x,
                             upper=upper,
                             lower=lower,
                             diagonal=diagonal)
    ns = calc_n(axes)
    assert ns['upper'] == upper * nx*(nx-1)//2
    assert ns['lower'] == lower * nx*(nx-1)//2
    assert ns['diagonal'] == diagonal * nx

    params = [paramnames_x, paramnames_y]
    all_params = paramnames_x + paramnames_y
    nu, nl, nd = 0, 0, 0
    for x in paramnames_x:
        for y in paramnames_y:
            if x == y:
                nd += 1
            elif all_params.index(x) < all_params.index(y):
                nl += 1
            elif all_params.index(x) > all_params.index(y):
                nu += 1
    fig, axes = make_2d_axes(params,
                             upper=upper,
                             lower=lower,
                             diagonal=diagonal)
    ns = calc_n(axes)
    assert ns['upper'] == upper * nu
    assert ns['lower'] == lower * nl
    assert ns['diagonal'] == diagonal * nd


@pytest.mark.parametrize('upper', [False, True])
@pytest.mark.parametrize('ticks', ['inner', 'outer', None])
def test_make_2d_axes_ticks(upper, ticks):
    xticks = [0.1, 0.4, 0.7]
    yticks = [0.2, 0.5, 0.8]
    paramnames = ["x0", "x1", "x2", "x3"]
    for k in paramnames:
        fig, axes = make_2d_axes(paramnames, upper=upper, ticks=ticks)
        axes[k][k].set_xticks(xticks)
        axes[k][k].set_yticks(yticks)
        for i, row in axes.iterrows():
            for j, ax in row.items():
                if ax is None:
                    break
                if i == k:
                    assert np.array_equal(yticks, ax.get_yticks())
                else:
                    assert not np.array_equal(yticks, ax.get_yticks())
                if j == k:
                    assert np.array_equal(xticks, ax.get_xticks())
                else:
                    assert not np.array_equal(xticks, ax.get_xticks())
    with pytest.raises(ValueError):
        make_2d_axes(paramnames, upper=upper, ticks='spam')


def test_make_2d_axes_ticks_error():
    with pytest.raises(ValueError):
        make_2d_axes(['a', 'b'], ticks='spam')


def test_2d_axes_limits():
    np.random.seed(0)
    paramnames = ['A', 'B', 'C', 'D']
    fig, axes = make_2d_axes(paramnames)
    for x in paramnames:
        for y in paramnames:
            a, b, c, d = np.random.rand(4)
            axes[x][y].set_xlim(a, b)
            for z in paramnames:
                assert axes[x][z].get_xlim() == (a, b)
                assert axes[z][x].get_ylim() == (a, b)

            axes[x][y].set_ylim(c, d)
            for z in paramnames:
                assert axes[y][z].get_xlim() == (c, d)
                assert axes[z][y].get_ylim() == (c, d)


@pytest.mark.parametrize('axesparams', [['A', 'B', 'C', 'D'],
                                        [['A', 'B', 'C', 'D'], ['A', 'B']],
                                        [['A', 'B'], ['A', 'B', 'C', 'D']]])
@pytest.mark.parametrize('params', [{'A': 0},
                                    {'A': 0, 'C': 0, 'E': 0},
                                    {'A': 0, 'C': [0, 0.5]}])
@pytest.mark.parametrize('upper', [True, False])
def test_2d_axes_axlines(axesparams, params, upper):
    kwargs = dict(c='k', ls='--', lw=0.5)
    fig, axes = make_2d_axes(axesparams, upper=upper)
    axes.axlines(params, **kwargs)


@pytest.mark.parametrize('axesparams', [['A', 'B', 'C', 'D'],
                                        [['A', 'B', 'C', 'D'], ['A', 'B']],
                                        [['A', 'B'], ['A', 'B', 'C', 'D']]])
@pytest.mark.parametrize('params', [{'A': (0, 0.1)},
                                    {'A': (0, 1), 'C': (0, 1), 'E': (0, 1)},
                                    {'A': (0, 1), 'C': [(-0.1, 0), (0.5, 1)]}])
@pytest.mark.parametrize('upper', [True, False])
def test_2d_axes_axspans(axesparams, params, upper):
    kwargs = dict(c='k', alpha=0.5)
    fig, axes = make_2d_axes(axesparams, upper=upper)
    axes.axspans(params, **kwargs)


@pytest.mark.parametrize('axesparams', [['A', 'B', 'C', 'D'],
                                        [['A', 'B', 'C', 'D'], ['A', 'B']],
                                        [['A', 'B'], ['A', 'B', 'C', 'D']]])
@pytest.mark.parametrize('params', [{'A': 0},
                                    {'A': 0, 'C': 0, 'E': 0},
                                    {'A': [0, 0.1], 'C': [0, 0.5]}])
@pytest.mark.parametrize('upper', [True, False])
def test_2d_axes_scatter(axesparams, params, upper):
    kwargs = dict(c='k', marker='*')
    fig, axes = make_2d_axes(axesparams, upper=upper)
    axes.scatter(params, **kwargs)


@pytest.mark.parametrize('plot_1d', [kde_plot_1d,
                                     skipif_no_fastkde(fastkde_plot_1d)])
def test_kde_plot_1d(plot_1d):
    fig, ax = plt.subplots()
    np.random.seed(0)
    data = np.random.randn(1000)

    # Check height
    line, = plot_1d(ax, data)
    assert isinstance(line, Line2D)
    assert line.get_ydata().max() <= 1

    # Check arguments are passed onward to underlying function
    line, = plot_1d(ax, data, color='r')
    assert line.get_color() == 'r'
    line, = plot_1d(ax, data, cmap=plt.cm.Blues)
    assert line.get_color() == plt.cm.Blues(0.68)

    # Check q
    plot_1d(ax, data, q='1sigma')
    plot_1d(ax, data, q=0)
    plot_1d(ax, data, q=1)
    plot_1d(ax, data, q=5)
    plot_1d(ax, data, q=10)
    plot_1d(ax, data, q=0.1)
    plot_1d(ax, data, q=0.9)
    plot_1d(ax, data, q=(0.1, 0.9))

    # Check iso-probability code
    line, fill = plot_1d(ax, data, facecolor=True)
    plot_1d(ax, data, facecolor=True, levels=[0.8, 0.6, 0.2])
    line, fill = plot_1d(ax, data, fc='blue', color='k', ec='r')
    assert np.all(fill[0].get_edgecolor()[0] == to_rgba('r'))
    assert (to_rgba(line[0].get_color()) == to_rgba('r'))
    line, fill = plot_1d(ax, data, fc=True, color='k', ec=None)
    assert len(fill[0].get_edgecolor()) == 0
    assert (to_rgba(line[0].get_color()) == to_rgba('k'))

    # Check levels
    with pytest.raises(ValueError):
        fig, ax = plt.subplots()
        plot_1d(ax, data, fc=True, levels=[0.68, 0.95])

    # Check xlim, Gaussian (i.e. limits reduced to relevant data range)
    fig, ax = plt.subplots()
    data = np.random.randn(1000) * 0.01 + 0.5
    plot_1d(ax, data)
    xmin, xmax = ax.get_xlim()
    assert xmin > 0.4
    assert xmax < 0.6
    # Check xlim, Uniform (i.e. data and limits span entire prior boundary)
    fig, ax = plt.subplots()
    data = np.random.uniform(size=1000)
    plot_1d(ax, data, q=0)
    xmin, xmax = ax.get_xlim()
    assert xmin <= 0
    assert xmax >= 1


@fastkde_mark_xfail
def test_fastkde_min_max():
    np.random.seed(0)
    data_x = np.random.randn(100)
    data_y = np.random.randn(100)
    xmin, xmax = -1, +1
    ymin, ymax = -1, +1
    _, ax = plt.subplots()
    line, = fastkde_plot_1d(ax, data_x, xmin=xmin)
    assert (line.get_xdata() >= xmin).all()

    _, ax = plt.subplots()
    line, = fastkde_plot_1d(ax, data_x, xmax=xmax)
    assert (line.get_xdata() <= xmax).all()

    _, ax = plt.subplots()
    line, = fastkde_plot_1d(ax, data_x, xmin=xmin, xmax=xmax)
    assert (line.get_xdata() >= xmin).all()
    assert (line.get_xdata() <= xmax).all()

    _, ax = plt.subplots()
    fastkde_contour_plot_2d(ax, data_x, data_y, xmin=xmin, ymin=ymin)
    assert ax.get_xlim()[0] >= xmin
    assert ax.get_ylim()[0] >= ymin

    _, ax = plt.subplots()
    fastkde_contour_plot_2d(ax, data_x, data_y, xmax=xmax, ymax=ymax)
    assert ax.get_xlim()[1] <= xmax
    assert ax.get_ylim()[1] <= ymax

    _, ax = plt.subplots()
    fastkde_contour_plot_2d(ax, data_x, data_y,
                            xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax)
    assert ax.get_xlim()[0] >= xmin
    assert ax.get_xlim()[1] <= xmax
    assert ax.get_ylim()[0] >= ymin
    assert ax.get_ylim()[1] <= ymax


def test_hist_plot_1d():
    fig, ax = plt.subplots()
    np.random.seed(0)
    data = np.random.randn(1000)

    # Check heights for histtype 'bar'
    bars = hist_plot_1d(ax, data, histtype='bar')[2]
    assert np.all([isinstance(b, Patch) for b in bars])
    assert max([b.get_height() for b in bars]) == 1.
    assert np.all(np.array([b.get_height() for b in bars]) <= 1.)

    # Check heights for histtype 'step'
    polygon, = hist_plot_1d(ax, data, histtype='step')[2]
    assert isinstance(polygon, Polygon)
    trans = polygon.get_transform() - ax.transData
    assert np.isclose(trans.transform(polygon.xy)[:, -1].max(), 1.,
                      rtol=1e-10, atol=1e-10)
    assert np.all(trans.transform(polygon.xy)[:, -1] <= 1.)

    # Check heights for histtype 'stepfilled'
    polygon, = hist_plot_1d(ax, data, histtype='stepfilled')[2]
    assert isinstance(polygon, Polygon)
    trans = polygon.get_transform() - ax.transData
    assert np.isclose(trans.transform(polygon.xy)[:, -1].max(), 1.,
                      rtol=1e-10, atol=1e-10)
    assert np.all(trans.transform(polygon.xy)[:, -1] <= 1.)

    # Check arguments are passed onward to underlying function
    bars = hist_plot_1d(ax, data, histtype='bar', color='r', alpha=0.5)[2]
    cc = ColorConverter.to_rgba('r', alpha=0.5)
    assert np.all([b.get_fc() == cc for b in bars])
    bars = hist_plot_1d(ax, data, histtype='bar', cmap=plt.cm.viridis,
                        alpha=0.5)[2]
    cc = ColorConverter.to_rgba(plt.cm.viridis(0.68), alpha=0.5)
    assert np.all([b.get_fc() == cc for b in bars])
    polygon, = hist_plot_1d(ax, data, histtype='step', color='r', alpha=0.5)[2]
    assert polygon.get_ec() == ColorConverter.to_rgba('r', alpha=0.5)
    polygon, = hist_plot_1d(ax, data, histtype='step', cmap=plt.cm.viridis,
                            color='r')[2]
    assert polygon.get_ec() == ColorConverter.to_rgba('r')


@astropy_mark_xfail
@pytest.mark.parametrize('bins', ['knuth', 'freedman', 'blocks'])
def test_astropyhist_plot_1d(bins):
    fig, ax = plt.subplots()
    np.random.seed(0)
    data = np.random.randn(100)
    with pytest.raises(ValueError):
        hist_plot_1d(ax, data, bins=bins)


@pytest.mark.parametrize('bins', ['fd', 'scott', 'sqrt'])
def test_hist_plot_1d_bins(bins):
    np.random.seed(0)
    num = 1000
    bound = 5
    x = np.random.uniform(-bound, +bound, num)
    w = stats.norm.pdf(x)
    fig, ax = plt.subplots()
    _, edges, _ = hist_plot_1d(ax, x, weights=w, bins=bins)
    assert 15 <= edges.size <= 30
    assert edges[+0] > x.min()
    assert edges[-1] < x.max()

    _, edges1, _ = hist_plot_1d(ax, x, weights=w, bins=bins, beta=1)
    _, edges2, _ = hist_plot_1d(ax, x, weights=w, bins=bins, beta=2)
    assert edges1.size > edges2.size > edges.size

    _, edgesr, _ = hist_plot_1d(ax, x, weights=w, bins=bins, range=(-3, 3))
    assert 10 <= edgesr.size <= edges.size
    assert edgesr[0] == -3
    assert edgesr[-1] == 3

    _, edges, _ = hist_plot_1d(ax, x, weights=w, bins=bins, range=None)
    assert 15 <= edges.size <= 30
    assert edges[+0] == pytest.approx(x.min(), abs=2*bound/num)
    assert edges[-1] == pytest.approx(x.max(), abs=2*bound/num)

    _, edges, _ = hist_plot_1d(ax, x, weights=None, bins=bins)
    assert 10 <= edges.size <= 35
    assert edges[+0] == pytest.approx(x.min(), abs=2*bound/num)
    assert edges[-1] == pytest.approx(x.max(), abs=2*bound/num)


def test_hist_plot_2d():
    fig, ax = plt.subplots()
    np.random.seed(0)
    data_x, data_y = np.random.randn(2, 1000)
    hist_plot_2d(ax, data_x, data_y)
    xmin, xmax = ax.get_xlim()
    ymin, ymax = ax.get_ylim()
    assert xmin > -5 and xmax < 5 and ymin > -5 and ymax < 5

    fig, ax = plt.subplots()
    data_x, data_y = np.random.uniform(-10, 10, (2, 1000))
    hist_plot_2d(ax, data_x, data_y)
    xmin, xmax = ax.get_xlim()
    ymin, ymax = ax.get_ylim()
    assert xmin > -10 and xmax < 10 and ymin > -10 and ymax < 10

    fig, ax = plt.subplots()
    data_x, data_y = np.random.uniform(-10, 10, (2, 1000))
    weights = np.exp(-(data_x**2 + data_y**2)/2)
    hist_plot_2d(ax, data_x, data_y, weights=weights, bins=30)
    xmin, xmax = ax.get_xlim()
    ymin, ymax = ax.get_ylim()
    assert xmin > -6 and xmax < 6 and ymin > -6 and ymax < 6

    fig, ax = plt.subplots()
    data_x, data_y = np.random.randn(2, 1000)
    hist_plot_2d(ax, data_x, data_y, levels=[0.95, 0.68], cmin=50)
    hist_plot_2d(ax, data_x, data_y, levels=[0.95, 0.68], cmax=50)


@pytest.mark.parametrize('plot_1d', [kde_plot_1d,
                                     skipif_no_fastkde(fastkde_plot_1d)])
@pytest.mark.parametrize('s', [1, 2])
def test_1d_density_kwarg(plot_1d, s):
    np.random.seed(0)
    x = np.random.normal(scale=s, size=2000)
    fig, ax = plt.subplots()

    # hist density = False:
    h = hist_plot_1d(ax, x, density=False,
                     bins=np.linspace(-5.5, 5.5, 12))[2]
    bar_height = h.get_children()[len(h.get_children()) // 2].get_height()
    assert bar_height == pytest.approx(1, rel=0.1)

    # kde density = False:
    k = plot_1d(ax, x, density=False)[0]
    f = interp1d(k.get_xdata(), k.get_ydata(), 'cubic', assume_sorted=True)
    kde_height = f(0)
    assert kde_height == pytest.approx(1, rel=0.1)

    # hist density = True:
    h = hist_plot_1d(ax, x, density=True,
                     bins=np.linspace(-5.5, 5.5, 12))[2]
    bar_height = h.get_children()[len(h.get_children()) // 2].get_height()
    assert bar_height == pytest.approx(erf(0.5 / np.sqrt(2) / s), rel=0.1)

    # kde density = True:
    k = plot_1d(ax, x, density=True)[0]
    f = interp1d(k.get_xdata(), k.get_ydata(), 'cubic', assume_sorted=True)
    kde_height = f(0)
    gauss_norm = 1 / np.sqrt(2 * np.pi * s**2)
    assert kde_height == pytest.approx(gauss_norm, rel=0.1)


@pytest.mark.parametrize('contour_plot_2d',
                         [kde_contour_plot_2d,
                          skipif_no_fastkde(fastkde_contour_plot_2d)])
def test_contour_plot_2d(contour_plot_2d):
    fig, ax = plt.subplots()
    np.random.seed(1)
    data_x = np.random.randn(1000)
    data_y = np.random.randn(1000)
    cf, ct = contour_plot_2d(ax, data_x, data_y)
    assert isinstance(cf, ContourSet)
    assert isinstance(ct, ContourSet)

    # Check levels
    with pytest.raises(ValueError):
        fig, ax = plt.subplots()
        contour_plot_2d(ax, data_x, data_y, levels=[0.68, 0.95])

    # Check unfilled
    cmap = basic_cmap('C2')
    fig, ax = plt.subplots()
    cf1, ct1 = contour_plot_2d(ax, data_x, data_y, facecolor='C2')
    cf2, ct2 = contour_plot_2d(ax, data_x, data_y, fc='None', cmap=cmap)
    # filled `contourf` and unfilled `contour` colors are the same:
    # tcolors deprecated in matplotlib 3.8
    cf1_tcolors = [tuple(rgba) for rgba in cf1.to_rgba(cf1.cvalues, cf1.alpha)]
    ct2_tcolors = [tuple(rgba) for rgba in ct2.to_rgba(ct2.cvalues, ct2.alpha)]

    assert cf1_tcolors[0] == ct2_tcolors[0]
    assert cf1_tcolors[1] == ct2_tcolors[1]
    cf, ct = contour_plot_2d(ax, data_x, data_y, edgecolor='C0')
    assert ct.colors == 'C0'
    cf, ct = contour_plot_2d(ax, data_x, data_y, ec='C0', cmap=plt.cm.Reds)
    assert cf.get_cmap() == plt.cm.Reds
    assert ct.colors == 'C0'
    fig, ax = plt.subplots()
    cf, ct = contour_plot_2d(ax, data_x, data_y, fc=None)
    assert cf is None
    assert ct.colors is None
    assert ct.get_cmap()(1.) == to_rgba('C0')
    cf, ct = contour_plot_2d(ax, data_x, data_y, fc=None, c='C3')
    assert cf is None
    assert ct.colors is None
    assert ct.get_cmap()(1.) == to_rgba('C3')
    cf, ct = contour_plot_2d(ax, data_x, data_y, fc=None, ec='C1')
    assert cf is None
    assert ct.colors == 'C1'
    cf, ct = contour_plot_2d(ax, data_x, data_y, fc=None, cmap=plt.cm.Reds)
    assert cf is None
    assert ct.colors is None
    assert ct.get_cmap() == plt.cm.Reds

    # Check limits, Gaussian (i.e. limits reduced to relevant data range)
    fig, ax = plt.subplots()
    data_x = np.random.randn(1000) * 0.01 + 0.5
    data_y = np.random.randn(1000) * 0.01 + 0.5
    contour_plot_2d(ax, data_x, data_y)
    xmin, xmax = ax.get_xlim()
    ymin, ymax = ax.get_ylim()
    assert xmin > 0.4
    assert xmax < 0.6
    assert ymin > 0.4
    assert ymax < 0.6
    # Check limits, Uniform (i.e. data & limits span entire prior boundary)
    fig, ax = plt.subplots()
    data_x = np.random.uniform(size=1000)
    data_y = np.random.uniform(size=1000)
    contour_plot_2d(ax, data_x, data_y, q=0)
    xmin, xmax = ax.get_xlim()
    ymin, ymax = ax.get_ylim()
    if contour_plot_2d is fastkde_contour_plot_2d:
        assert xmin <= 0
        assert xmax >= 1
        assert ymin <= 0
        assert ymax >= 1
    elif contour_plot_2d is kde_contour_plot_2d:
        assert xmin == pytest.approx(0, abs=0.01)
        assert xmax == pytest.approx(1, abs=0.01)
        assert ymin == pytest.approx(0, abs=0.01)
        assert ymax == pytest.approx(1, abs=0.01)


@pytest.mark.parametrize('plot_1d', [kde_plot_1d, hist_plot_1d])
def test_q_1d(plot_1d):
    np.random.seed(42)
    d = np.random.randn(1000)
    fig, ax = plt.subplots()

    # check that axis limits match data limits for single plot
    p = plot_1d(ax, d, q=1)  # int input for q
    p = p[1] if plot_1d == hist_plot_1d else p[0].get_data()[0]
    assert ax.get_xlim() == pytest.approx((-1, 1), rel=0.1)
    assert (p.min(), p.max()) == pytest.approx((-1, 1), rel=0.1)

    # check that axis limits grow with new, larger plot
    p = plot_1d(ax, d, q="2sigma")  # str input for q
    p = p[1] if plot_1d == hist_plot_1d else p[0].get_data()[0]
    assert ax.get_xlim() == pytest.approx((-2, 2), rel=0.15)
    assert (p.min(), p.max()) == pytest.approx((-2, 2), rel=0.1)

    # check that axis limits do not shrink with new, smaller plot
    p = plot_1d(ax, d, q=(0.025, 0.84))  # tuple input for q
    p = p[1] if plot_1d == hist_plot_1d else p[0].get_data()[0]
    assert ax.get_xlim() == pytest.approx((-2, 2), rel=0.15)
    assert (p.min(), p.max()) == pytest.approx((-2, 1), rel=0.1)


@pytest.mark.parametrize('plot_2d', [kde_contour_plot_2d,
                                     hist_plot_2d,
                                     scatter_plot_2d])
def test_q_2d(plot_2d):
    np.random.seed(42)
    d = np.random.randn(2, 1000)

    def get_data_from_plot(plot_2d, p):
        if plot_2d == kde_contour_plot_2d:
            x, y = p[0].allsegs[0][0].T
        elif plot_2d == hist_plot_2d:
            x = p.get_coordinates()[0, :, 0]
            y = p.get_coordinates()[:, 0, 1]
        elif plot_2d == scatter_plot_2d:
            x, y = p[0].get_xydata().T
        return x, y

    def expected_limits(plot_2d, x, y):
        if plot_2d == scatter_plot_2d:
            xpad = (x.max()-x.min()) * plt.rcParams['axes.xmargin']
            ypad = (y.max()-y.min()) * plt.rcParams['axes.ymargin']
            return (x.min()-xpad, x.max()+xpad), (y.min()-ypad, y.max()+ypad)
        return (x.min(), x.max()), (y.min(), y.max())

    fig, ax = plt.subplots()
    # check that axis limits match data limits for single plot
    p = plot_2d(ax, d[0], d[1], q=1)  # int input for q
    x, y = get_data_from_plot(plot_2d, p)
    xlim, ylim = expected_limits(plot_2d, x, y)
    assert ax.get_xlim() == pytest.approx((-1, 1), rel=0.2)
    assert ax.get_ylim() == pytest.approx((-1, 1), rel=0.2)
    assert ax.get_xlim() == pytest.approx(xlim, rel=1e-15)
    assert ax.get_ylim() == pytest.approx(ylim, rel=1e-15)

    # check that axis limits grow with new, larger plot
    p = plot_2d(ax, d[0], d[1], q="2sigma")  # str input for q
    x, y = get_data_from_plot(plot_2d, p)
    xlim, ylim = expected_limits(plot_2d, x, y)
    assert ax.get_xlim() == pytest.approx((-2, 2), rel=0.2)
    assert ax.get_ylim() == pytest.approx((-2, 2), rel=0.2)
    assert ax.get_xlim() == pytest.approx(xlim, rel=1e-15)
    assert ax.get_ylim() == pytest.approx(ylim, rel=1e-15)

    # check that axis limits do not shrink with new, smaller plot
    p = plot_2d(ax, d[0], d[1], q=(0.025, 0.84))  # tuple input for q
    x, y = get_data_from_plot(plot_2d, p)
    assert ax.get_xlim() == pytest.approx((-2, 2), rel=0.2)
    assert ax.get_ylim() == pytest.approx((-2, 2), rel=0.2)
    assert (x.min(), x.max()) == pytest.approx((-2, 1), rel=0.2)
    assert (y.min(), y.max()) == pytest.approx((-2, 1), rel=0.2)


def test_kde_plot_nplot():
    fig, ax = plt.subplots()
    np.random.seed(0)
    data = np.random.randn(1000)
    line, = kde_plot_1d(ax, data, ncompress=1000, nplot_1d=200)
    assert line.get_xdata().size == 200

    fig, ax = plt.subplots()
    np.random.seed(0)
    data_x = np.random.randn(1000)
    data_y = np.random.randn(1000)
    kde_contour_plot_2d(ax, data_x, data_y, ncompress=1000, nplot_2d=900)


@pytest.mark.parametrize('contour_plot_2d',
                         [kde_contour_plot_2d,
                          skipif_no_fastkde(fastkde_contour_plot_2d)])
@pytest.mark.parametrize('levels', [[0.9],
                                    [0.9, 0.6],
                                    [0.9, 0.6, 0.3],
                                    [0.9, 0.7, 0.5, 0.3]])
def test_contour_plot_2d_levels(contour_plot_2d, levels):
    np.random.seed(42)
    x = np.random.randn(1000)
    y = np.random.randn(1000)
    cmap = plt.cm.viridis

    fig, (ax1, ax2) = plt.subplots(2)
    contour_plot_2d(ax1, x, y, levels=levels, cmap=cmap)
    contour_plot_2d(ax2, x, y, levels=levels, cmap=cmap, fc=None)

    # assert that color between filled and unfilled contours matches
    if version.parse(matplotlib.__version__) >= version.parse('3.8.0'):
        color1 = ax1.collections[0].get_facecolor()  # filled face color
        color2 = ax2.collections[0].get_edgecolor()  # unfilled line color
        # first level
        assert_array_equal(color1[0], color2[0])
        # last level
        assert_array_equal(color1[len(levels)-1], color2[len(levels)-1])
    else:
        # first level
        color1 = ax1.collections[0].get_facecolor()  # filled face color
        color2 = ax2.collections[0].get_edgecolor()  # unfilled line color
        assert_array_equal(color1, color2)
        # last level
        color1 = ax1.collections[len(levels)-1].get_facecolor()
        color2 = ax2.collections[len(levels)-1].get_edgecolor()
        assert_array_equal(color1, color2)


def test_scatter_plot_2d():
    fig, ax = plt.subplots()
    np.random.seed(2)
    data_x = np.random.randn(100)
    data_y = np.random.randn(100)
    lines, = scatter_plot_2d(ax, data_x, data_y)
    assert isinstance(lines, Line2D)

    fig, ax = plt.subplots()
    points, = scatter_plot_2d(ax, data_x, data_y, color='C0', lw=1)
    assert (points.get_color() == 'C0')
    points, = scatter_plot_2d(ax, data_x, data_y, cmap=plt.cm.viridis)
    assert (points.get_color() == plt.cm.viridis(0.68))
    points, = scatter_plot_2d(ax, data_x, data_y, c='C0', fc='C1', ec='C2')
    assert (points.get_color() == 'C0')
    assert (points.get_markerfacecolor() == 'C1')
    assert (points.get_markeredgecolor() == 'C2')

    # Check that q is ignored
    fig, ax = plt.subplots()
    scatter_plot_2d(ax, data_x, data_y, q=0)


def test_make_axes_logscale():
    # 1d
    fig, axes = make_1d_axes(['x0', 'x1', 'x2', 'x3'], logx=['x1', 'x3'])
    assert axes.loc['x0'].get_xscale() == 'linear'
    assert axes.loc['x1'].get_xscale() == 'log'
    assert axes.loc['x2'].get_xscale() == 'linear'
    assert axes.loc['x3'].get_xscale() == 'log'

    # 2d, logx only
    fig, axes = make_2d_axes(['x0', 'x1', 'x2', 'x3'],
                             logx=['x1', 'x3'])
    for y, rows in axes.iterrows():
        for x, ax in rows.items():
            if x in ['x0', 'x2']:
                assert ax.get_xscale() == 'linear'
            else:
                assert ax.get_xscale() == 'log'
            assert ax.get_yscale() == 'linear'
            with catch_warnings():
                filterwarnings('error', category=UserWarning,
                               message="Attempt to set non-positive")
                ax.set_ylim(-1, 1)

    # 2d, logy only
    fig, axes = make_2d_axes(['x0', 'x1', 'x2', 'x3'],
                             logy=['x1', 'x3'])
    for y, rows in axes.iterrows():
        for x, ax in rows.items():
            assert ax.get_xscale() == 'linear'
            with catch_warnings():
                filterwarnings('error', category=UserWarning,
                               message="Attempt to set non-positive")
                ax.set_xlim(-1, 1)
            if y in ['x0', 'x2']:
                assert ax.get_yscale() == 'linear'
            else:
                assert ax.get_yscale() == 'log'

    # 2d, logx and logy
    fig, axes = make_2d_axes(['x0', 'x1', 'x2', 'x3'],
                             logx=['x1', 'x3'],
                             logy=['x1', 'x3'])
    for y, rows in axes.iterrows():
        for x, ax in rows.items():
            if x in ['x0', 'x2']:
                assert ax.get_xscale() == 'linear'
            else:
                assert ax.get_xscale() == 'log'
            if y in ['x0', 'x2']:
                assert ax.get_yscale() == 'linear'
            else:
                assert ax.get_yscale() == 'log'


@pytest.mark.parametrize('plot_1d', [kde_plot_1d,
                                     skipif_no_fastkde(fastkde_plot_1d),
                                     hist_plot_1d])
def test_logscale_1d(plot_1d):
    np.random.seed(42)
    logdata = np.random.randn(1000)
    data = 10**logdata

    fig, ax = plt.subplots()
    ax.set_xscale('log')
    p = plot_1d(ax, data)
    if 'kde' in plot_1d.__name__:
        amax = abs(np.log10(p[0].get_xdata()[np.argmax(p[0].get_ydata())]))
    else:
        amax = abs(np.log10(p[1][np.argmax(p[0])]))
    assert amax < 0.5


@pytest.mark.parametrize('b', ['scott', 20, np.logspace(-5, 5, 20)])
def test_logscale_hist_kwargs(b):
    np.random.seed(42)
    logdata = np.random.randn(1000)
    data = 10**logdata

    fig, ax = plt.subplots()
    ax.set_xscale('log')
    h, edges, _ = hist_plot_1d(ax, data, bins=b)
    amax = abs(np.log10(edges[np.argmax(h)]))
    assert amax < 0.5
    assert edges[0] < 1e-3
    assert edges[-1] > 1e3
    h, edges, _ = hist_plot_1d(ax, data, bins=b, range=None)
    amax = abs(np.log10(edges[np.argmax(h)]))
    assert amax < 0.5
    assert edges[0] < 1e-3
    assert edges[-1] > 1e3
    h, edges, _ = hist_plot_1d(ax, data, bins=b, range=(1e-3, 1e3))
    amax = abs(np.log10(edges[np.argmax(h)]))
    assert amax < 0.5
    if isinstance(b, (int, str)):
        # edges are trimmed according to range
        assert edges[0] == 1e-3
        assert edges[-1] == 1e3
    else:
        # edges passed directly to bins are not trimmed according to range
        assert edges[0] == b[0]
        assert edges[-1] == b[-1]


@pytest.mark.parametrize('plot_2d',
                         [kde_contour_plot_2d,
                          skipif_no_fastkde(fastkde_contour_plot_2d),
                          hist_plot_2d, scatter_plot_2d])
def test_logscale_2d(plot_2d):
    np.random.seed(0)
    logx = np.random.randn(1000)
    logy = np.random.randn(1000)
    x = 10**logx
    y = 10**logy

    # logx
    fig, ax = plt.subplots()
    ax.set_xscale('log')
    p = plot_2d(ax, x, logy)
    if 'kde' in plot_2d.__name__:
        if version.parse(matplotlib.__version__) >= version.parse('3.8.0'):
            xmax, ymax = p[0].get_paths()[1].vertices[0].T
        else:
            xmax, ymax = p[0].allsegs[1][0].T
        xmax = np.mean(np.log10(xmax))
        ymax = np.mean(ymax)
    elif 'hist' in plot_2d.__name__:
        c = p.get_coordinates()
        c = (c[:-1, :] + c[1:, :]) / 2
        c = (c[:, :-1] + c[:, 1:]) / 2
        c = c.reshape((-1, 2))
        xmax = abs(np.log10(c[np.argmax(p.get_array())][0]))
        ymax = abs(c[np.argmax(p.get_array())][1])
    else:
        xmax = np.mean(np.log10(p[0].get_xdata()))
        ymax = np.mean(p[0].get_ydata())
    assert xmax < 0.5
    assert ymax < 0.5

    # logy
    fig, ax = plt.subplots()
    ax.set_yscale('log')
    p = plot_2d(ax, logx, y)
    if 'kde' in plot_2d.__name__:
        if version.parse(matplotlib.__version__) >= version.parse('3.8.0'):
            xmax, ymax = p[0].get_paths()[1].vertices[0].T
        else:
            xmax, ymax = p[0].allsegs[1][0].T
        xmax = np.mean(xmax)
        ymax = np.mean(np.log10(ymax))
    elif 'hist' in plot_2d.__name__:
        c = p.get_coordinates()
        c = (c[:-1, :] + c[1:, :]) / 2
        c = (c[:, :-1] + c[:, 1:]) / 2
        c = c.reshape((-1, 2))
        xmax = abs(c[np.argmax(p.get_array())][0])
        ymax = abs(np.log10(c[np.argmax(p.get_array())][1]))
    else:
        xmax = np.mean(p[0].get_xdata())
        ymax = np.mean(np.log10(p[0].get_ydata()))
    assert xmax < 0.5
    assert ymax < 0.5

    # logx and logy
    fig, ax = plt.subplots()
    ax.set_xscale('log')
    ax.set_yscale('log')
    p = plot_2d(ax, x, y)
    if 'kde' in plot_2d.__name__:
        if version.parse(matplotlib.__version__) >= version.parse('3.8.0'):
            xmax, ymax = p[0].get_paths()[1].vertices[0].T
        else:
            xmax, ymax = p[0].allsegs[1][0].T
        xmax = np.mean(np.log10(xmax))
        ymax = np.mean(np.log10(ymax))
    elif 'hist' in plot_2d.__name__:
        c = p.get_coordinates()
        c = (c[:-1, :] + c[1:, :]) / 2
        c = (c[:, :-1] + c[:, 1:]) / 2
        c = c.reshape((-1, 2))
        xmax = abs(np.log10(c[np.argmax(p.get_array())][0]))
        ymax = abs(np.log10(c[np.argmax(p.get_array())][1]))
    else:
        xmax = np.mean(np.log10(p[0].get_xdata()))
        ymax = np.mean(np.log10(p[0].get_ydata()))
    assert xmax < 0.5
    assert ymax < 0.5


@pytest.mark.parametrize('sigmas', [(1, '1sigma', 0.682689492137086),
                                    (2, '2sigma', 0.954499736103642),
                                    (3, '3sigma', 0.997300203936740),
                                    (4, '4sigma', 0.999936657516334),
                                    (5, '5sigma', 0.999999426696856)])
def test_quantile_plot_interval_str(sigmas):
    qi1, qi2 = quantile_plot_interval(q=sigmas[0])
    qs1, qs2 = quantile_plot_interval(q=sigmas[1])
    assert qi1 == pytest.approx(0.5 - sigmas[2] / 2)
    assert qi2 == pytest.approx(0.5 + sigmas[2] / 2)
    assert qs1 == pytest.approx(0.5 - sigmas[2] / 2)
    assert qs2 == pytest.approx(0.5 + sigmas[2] / 2)


@pytest.mark.parametrize('floats', [0, 0.1, 0.9])
def test_quantile_plot_interval_float(floats):
    q1, q2 = quantile_plot_interval(q=floats)
    assert q1 == min(floats, 1 - floats)
    assert q2 == max(floats, 1 - floats)


@pytest.mark.parametrize('q1, q2', [(0, 1), (0.1, 0.9), (0, 0.9), (0.1, 1)])
def test_quantile_plot_interval_tuple(q1, q2):
    _q1, _q2 = quantile_plot_interval(q=(q1, q2))
    assert _q1 == q1
    assert _q2 == q2


@pytest.mark.parametrize('color', ['C0', 'k', 'gold', '#00FFFF',
                                   (1.0, 1.0, 0.0, 1.0)])
def test_basic_cmap(color):
    cmap = basic_cmap(color)

    # Check that the basic cmap is reversible
    cmap.reversed()



def replace_inner_function(outer, new_inner):
    """Replace a nested function code object used by outer with new_inner.

    The replacement new_inner must use the same name and must at most use the
    same closures as the original.

    Code taken from https://stackoverflow.com/a/27550237
    """
    # find original code object so we can validate the closures match
    ocode = outer.__code__
    function, code = type(outer), type(ocode)
    iname = new_inner.co_name
    orig_inner = next(
        const for const in ocode.co_consts
        if isinstance(const, code) and const.co_name == iname)

    # you can ignore later closures, but since they are matched by position
    # the new sequence must match the start of the old.
    assert (orig_inner.co_freevars[:len(new_inner.co_freevars)] ==
            new_inner.co_freevars), 'New closures must match originals'

    # replace the code object for the inner function
    new_consts = tuple(
        new_inner if const is orig_inner else const
        for const in outer.__code__.co_consts)

    # create a new code object with the new constants
    ncode = ocode.replace(co_consts=new_consts)

    # and a new function object using the updated code object
    return function(
        ncode, outer.__globals__, outer.__name__,
        outer.__defaults__, outer.__closure__
    )



"""Read NestedSamples from Nested_Fit chains."""
import os
import numpy as np
from anesthetic.read.getdist import read_paramnames
from anesthetic.samples import NestedSamples


def read_nestedfit(root, *args, **kwargs):
    """Read Nested_Fit chain files.

    Parameters
    ----------
    root : str
        root specify the directory only, no specific roots,
        The files read files are ``nf_output_points.txt``
        and ``nf_output_diag.txt``.

    """
    dead_file = os.path.join(root, 'nf_output_points.txt')
    birth_file = os.path.join(root, 'nf_output_diag.dat')
    data_dead = np.loadtxt(dead_file)
    data_birth = np.loadtxt(birth_file)
    weight, logL, data = np.split(data_dead, [1, 2], axis=1)
    logL_birth = data_birth[:, 0]
    root_getdist = os.path.join(root, 'nf_output_points')
    columns, labels = read_paramnames(root_getdist)
    # No specific labeling is implemented in nested_fit
    labels = columns
    columns = kwargs.pop('columns', columns)
    labels = kwargs.pop('labels', labels)
    kwargs['label'] = kwargs.get('label', os.path.basename(root))

    return NestedSamples(data=data, columns=columns,
                         logL=logL, logL_birth=logL_birth,
                         labels=labels, *args, **kwargs)
